/*  
    Copyright 2007, Francois Schiettekatte

    This file is part of Corteo.

    Corteo is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Corteo is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Corteo.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <float.h>

#include "constants.h"
#include "corteoindex.h"
#include "corteomatrix.h"
#include "corteoutil.h"
#include "randomx.h"
#include "techniques.h"

#ifdef MAX_THREAD
  #include <pthread.h>
  pthread_mutex_t mut;
#endif

#define MAXLAYERS  20   // number of target layers; absorber is an extra layer
#define MAXELEMENT 10   // number of elements in each layer
#define NUMDET     2    // number of detectors

// prime numbers for lists length
#define MAXLOGLIST 104729
#define MAXAZILIST 72211
#define MAXRANLIST 1000003

// inverf list loaded from file, fixed length
#define MAXERFLIST 79999

// global variables (faster than passing arguments)

// layer parameters
unsigned int numberLayers;                   // number of layers
unsigned int collisionLayer;                // index of the layer where the collision will occur
unsigned int numberElements[MAXLAYERS+1];   // number of elements in each layer

float       layerThickness[MAXLAYERS+1];    // thickness of each layer [A]
float       layerTop[MAXLAYERS+1];          // depth of the top of each layer [A]
float       layerBottom[MAXLAYERS+1];       // depth of the bottom of each layer [A]
float       layerDensity[MAXLAYERS+1];      // density of each layer [A^{-3}]
float       interatomicDist[MAXLAYERS+1];     // (3/(4 pi density))^(1/3) [A]

float       elementFraction[MAXLAYERS+1][MAXELEMENT];    // relative concentration of element in layer
float       elementFractionSum[MAXLAYERS+1][MAXELEMENT]; // cumulative concentration of elements in layer

// rather than calling lengthy functions, corteo uses large lists of precalculated values for certain functions
float       randomlist[MAXRANLIST];     // list of evenly distributed but randomly ordered values between 0 and 1
float       sqrtrandomlist[MAXRANLIST]; // sqrt of randomlist
float       sqrtloglist[MAXLOGLIST];    // list of evenly distributed but randomly ordered values of sqrt of -log of 1/MAXLOGLIST to 1
float       sqrtloglist1[MAXLOGLIST];   // 1/sqrtloglist
float       sinAzimAngle[MAXAZILIST];   // list cos and sin components of angles...
float       cosAzimAngle[MAXAZILIST];   //   ...this angle are evenly distributed but randomly ordered between 0 and 2*PI
float       erflist[MAXERFLIST];        // list of evenly distributed but randomly ordered values of inverse erf of -1+1/MAXERFLIST to 1-1/MAXERFLIST

float       localmatrix[DIME*DIMS];    

float       *hsp[MAXLAYERS+1];    // hydrogen stopping power table in each layer

// initial beam energy, minimum energy to follow a projectile (eV) 
float E0, minE0;

// energy difference discrimination for coincidence measurements (eV)
float EdiffDiscr;

// sqrt of mean free path ([A]^0.5)
float sqrtMeanFreePath[MAXLAYERS+1];

// energy calibration, detector energy resolution, number of channels in spectrum
float zero, slope, invSlope, quad, resol;
unsigned int nchannels;

// Factor by which the histogram is mltiplied: density*thickness*solidAngle/number_of_detected_particle, with  
//   number_of_detected_particle = numProjectile*solidAngle_virtual/cone_solid_angle, 
//   solidAngle_virtual = solidAngle*Wf*Hf, and cone_solid_angle = 2.*PI*(1.0f-cosAngSpread0)  
// The spectrum then needs (externally) to be multipllied by the number of incident ions q and the concentration
// of the detected element in the layer. See documentation for a more detailed explanation.
float factorNtOmegaByNumdetpart;

// projectile incident directional cosines
float l0, m0, n0;

// spot size (sigma of a gaussian distribution)
float ySpotSize, zSpotSize;

float massRatio; // for kinematic factor of mauy

// counters
unsigned long detectedParticles = 0, exitBottom = 0, aboveCutoff = 0, exitUndetected = 0, badRescale=0, 
              exitUndetectedFinal = 0, exitTop = 0, exitMinEnergy = 0, badKinematics=0;

// coordinates of points 1 and 3 in detector reference frame
float       p1yp[NUMDET], p1zp[NUMDET];
float       p3yp[NUMDET], p3zp[NUMDET]; 

// coordinates of detector center, factor d in plane equation
float       pcx[NUMDET], pcy[NUMDET], pcz[NUMDET], d[NUMDET];
float       pcnormx[NUMDET], pcnormy[NUMDET], pcnormz[NUMDET];

// components of normal vectors xp, yp, zp defining detectors reference frame 
float       xpx[NUMDET], xpy[NUMDET], xpz[NUMDET];
float       ypx[NUMDET], ypy[NUMDET], ypz[NUMDET]; 
float       zpx[NUMDET], zpy[NUMDET], zpz[NUMDET];

// variables indicating if detector is an ellipse
int         isEllipse[NUMDET]; 

// inverse of detector half-size in each direction, annulus radius squared
// (in case of ellipse, this is the inverse of major and minor axis, a & b)
float       ainv[NUMDET], binv[NUMDET], r2annular[NUMDET]; 

// factor in width (W: parallel to p3-p2) and height (H: parallel to p2-p1) 
// by which a virtual detector is larger than the real detector
float       Wf, Hf;

// angle of the cone around detector axis in which scattered event are generated
float       cosAngSpread0; 

// cos of the minimum angle beyond which collision is assumed to have missed
// (required because collision near 0 degrees contribute an enormous cross section in RBS)
float       cutoffCosthetaLab;

// consider (!=0) or do not consider (==0) cross section screening
unsigned int computeScreen;

// simulation type
int simulationType;

// follow (!=0) or do not follow (==0) recoils during implantation simulation
int followRecoils;

// writeDetails:
// ==0: no details
// ==1: write to corteo.details FOR EACH ION the energy, location and direction after main collision and at target exit
// ==2: write to corteo.angle FOR EACH ION the collision depth and angle with detector axis at target surface
// ==4: write to corteo.detec FOR EACH DECTECT ION its energy, intesection point with detector, directional cosines, and cross-section
// (values can be OR'd)
unsigned int  writeDetails;  

FILE *ofpdetails;
FILE *ofangle;
FILE *ofdetect;

// absorber thickness, 0 is no absorber
float       absorberThick;

// parameters for Chu straggling correction
float Chu[100][4];


// thread-related parameters
typedef struct {
  int id;
  unsigned long numProjectile;
  unsigned long totalnumbercoll;
  float * histogram;
} thread_params;



// ion parameters
typedef struct {
  unsigned int Z;
  float mass; 

  float a[MAXLAYERS+1][MAXELEMENT];            // screening length
  float f[MAXLAYERS+1][MAXELEMENT];            // reduced energy conversion factor
  float mr[MAXLAYERS+1][MAXELEMENT];           // mass ratio
  float K[MAXLAYERS+1][MAXELEMENT];            // kinematic factor
  float smax[MAXLAYERS+1][MAXELEMENT];         // maximum reduced impact parameter

  // pointers to tables of cos and sin of scattering angle
  float *CosScat[MAXLAYERS+1][MAXELEMENT];     
  float *SinScat[MAXLAYERS+1][MAXELEMENT];

  float  *sp[MAXLAYERS+1];        // stopping power table (computed according to Bragg's rule with compound correction)
  float  *sqrtStrag[MAXLAYERS+1]; // sqrt of energy straggling * sqrt(2) (computed according to Bragg's rule)

  int tablesComputed;

} ion_params;

ion_params beam;   // beam parameters
ion_params recoil; // recoil parameters 

// pointer to element parameters (implantation following recoils, same address for same Z, mass)
ion_params *element[MAXLAYERS+1][MAXELEMENT]; 

float *dispHist[NUMBERELEMENTS];  // pointer to displacements profile per element (only useful ones are allocated)
float *implHist[NUMBERELEMENTS];  // pointer to implantation profile per element (only useful ones are allocated)

// displacement and binding energy (for implantation, when following recoils)
// these are declatated outside ion_params structure because they can be different for different layers
// while the rest of the structure is identical for elements with same Z and mass whatever the layer is
float Ed[MAXLAYERS][MAXELEMENT]; 
float Eb[MAXLAYERS][MAXELEMENT]; 



// fill the table of cos and sin of the scattering angle in the lab frame given a mass ratio
void fillCosSinTable(float * cosTable, float * sinTable, float mr) {

  double sin2thetaby2, costheta, costhetaLab, sinthetaLab;
  unsigned long i, j;

  // invert matrix elements
  for(i=0; i<DIME; i++) 
    for(j=0; j<DIMS; j++) 
      localmatrix[i*DIMS+j] = Matrix(j*DIME+i);

  // compute scattering angle components
  for(i=0; i<DIME*DIMS; i++) {
    sin2thetaby2 = localmatrix[i];
    costheta = 1.-2.*sin2thetaby2;

    if(costheta==-1.0 && mr==1.0) {
      costhetaLab = 0.0;  // peculiar case of head-on collision of identical masses
      sinthetaLab = 1.0;
    } else {
      costhetaLab = (costheta+mr)/sqrt(1.+2.*mr*costheta+mr*mr);
      sinthetaLab = sqrt(1.-costhetaLab*costhetaLab);
    }
    cosTable[i] = d2f(costhetaLab);
    sinTable[i] = d2f(sinthetaLab);
  }
}


// load a list of inverse error function erfinv(x) values that contains MAXERFLIST elements
// the list is such that x is uniformly distributed between -1 and 1 excluding these boundaries
// i.e. x = -1+dx, -1+2dx, ...1-2dx, 1-dx, with 1/2dx = MAXERFLIST
// return 0 on error, 1 otherwise
int loadErfInv() {
  unsigned int k;
  char erfval[1000];
  FILE * ifp;

  ifp = fopen("data/erfinv.dat", "r");
  if(ifp==NULL) {
    fprintf(stderr,"Can't open data/erdinv.dat ");
    return 0;
  }

  for(k=0; k<MAXERFLIST; k++) {
    fscanf(ifp,"%s", erfval);
    erflist[k] = a2f(erfval);
  }
  fscanf(ifp,"%s", erfval);
  fclose(ifp);

  // control value at the end should be the number of elements in the list
  if(atoi(erfval)!=MAXERFLIST) return 0; 
  return 1;
}


// load Chu straggling correction parameters
// return 0 on error, 1 otherwise
int loadChu() {
  unsigned int k, l, Z;
  char Chuval[1000];
  FILE * ifp;

  ifp = fopen("data/chu.dat", "r");
  if(ifp==NULL) {
    fprintf(stderr,"Can't open data/chu.dat ");
    return 0;
  }

  ignoreline(ifp); // skip header line
  for(k=0; k<92; k++) {
    fscanf(ifp,"%u", &Z);
    for(l=0; l<4; l++) {
      fscanf(ifp,"%s", Chuval);
      Chu[Z][l] = a2f(Chuval);
    }
    if(Z!=k+2) // no data for Z=0,1, table starts at Z=2
      return 0;
  }
  fclose(ifp);
  return 1;
}

// - load dE/dx data for projectile particleZ in each element present in layer ilayer; 
//     - file is presumed to be in the right format, written by generateDedxTable with the right parameters
//     - control data should be wrong if something is wrong
// - compute stopping accoring to Bragg's rule;
// - result is stored in spp[]
// - return the adress of allocated memory if successfull; 
// - return NULL on error: stopping already set, can't open file or bad control data
// IMPORTANT: stopping for the most abundant isotope (energy for index calculation must be corrected for mass)
float * fillDedxTable(unsigned int ilayer, unsigned int projectileZ, float projectileMass, float dens, float compoundCorr) {
  char filename[1000];
  unsigned int i, k, elemZ;
  float f, buf[DIMD+1];
  FILE *ifp;
  char *mode = "rb";

  float * spp = calloc(DIMD,sizeof(float));  // allocate while initializing to 0
  if(spp == NULL) {
    fprintf(stderr,"Stopping power allocation error");
    return NULL;
  }

  sprintf(filename, "data/%u.asp",projectileZ);

  for(i=0; i<numberElements[ilayer]; i++) {
    elemZ = element[ilayer][i]->Z;
    
    // open file and move to record position
    ifp = fopen(filename, mode);
    if(ifp==NULL) {
      fprintf(stderr,"Can't open %s ", filename);
      return NULL;
  }
    fseek(ifp,sizeof(float)*(DIMD+1)*(elemZ-1),1);
    fread((void*)buf, sizeof(float), DIMD+1, ifp); // read dedx data plus control data
    fclose(ifp);
    
    if(buf[DIMD]!=elemZ) { // each record must end with the element number as a control data
      fprintf(stderr, "Error reading stopping power for particle Z=%u in material Z=%u\n",projectileZ, elemZ);
      return NULL; 
    }

    for(k=0; k<DIMD; k++) {
      // electronic stopping of index k; factor 10 is convertion from eV/(1E15 at/cm2) to eV/(at/A2), density is in at/A3
      spp[k]+=buf[k]*elementFraction[ilayer][i]*10.0f*dens*compoundCorr; // Bragg's rule
    }
  }

  // compound correction according to Zeigler & Manoyan NIMB35(1998)215, Eq.(16) (error in Eq. 14)
  if(compoundCorr!=1.0f)
    for(k=0; k<DIMD; k++) {
      f = d2f( 1.0/(1.0+exp( 1.48*( sqrt(2.*Dval(k)/projectileMass/25e3) -7.0) )) );
      spp[k]*=f*(compoundCorr-1.0f)+1.0f;
    }
  return spp;
}

// compute energy straggling of projectile (to be multiplied by distance in Angstroems during calculations)
// - compute straggling accoring to Bragg's rule;
// - compute for most abundant isotope mass
// - when stragType 
//          = 0: straggling = 0
//          = 1: Bohr's straggling
//          = 2: Chu correction
//          = 3: Chu + Wang correction
// - return the allocated table or NULL
// IMPORTANT: stopping for the most abundant isotope (energy for index calculation must be corrected for mass)
float * fillStragglingTable(unsigned int ilayer, unsigned int projectileZ, float stopping[], float hstopping[], int stragType) {
  unsigned int i, targetZ;
  unsigned long ii, k;
  double chargestate, E, EE, S;
  double omegaBohr2, chucorr, epsi, gamma, extra, str, sstrag; 
  double MAImass = MostAbundantIsotope[projectileZ];

  float * strag = calloc(DIMD,sizeof(float));  // allocate while initializing to 0
  if(strag == NULL) {
    fprintf(stderr,"Stopping power allocation error");
    return NULL;
  }

  for(k=0; k<DIMD; k++) {
    sstrag=0.;

    for(i=0; i<numberElements[ilayer]; i++) {
      targetZ = element[ilayer][i]->Z;

      S = stopping[k];
      E = Dval(k); // energy of index k

      // effective charge state squared is S(H)/S(ion) Z_ion^2 for S at same speed
      ii = Dindex(d2f(E/MAImass));
      chargestate = S/(hstopping[ii]*projectileZ*projectileZ);

      omegaBohr2 = 4.*PI*projectileZ*projectileZ*E2*E2*targetZ*layerDensity[ilayer];

      EE = E/MAImass/1.e6;
      chucorr = 1./(1.+Chu[targetZ][0]*pow(EE,Chu[targetZ][1])+Chu[targetZ][2]*pow(EE,Chu[targetZ][3]));

      epsi = EE/((double)projectileZ*sqrt((double)projectileZ*(double)targetZ)); 

      if(projectileZ==1)
        gamma = 2.522*(1.-exp(-1.040*EE));
      else
        gamma = 0.3931*(1.-exp(-3.812*epsi));

      if(projectileZ==1) 
        extra = 0.1955*gamma/(pow(EE-0.6941,2)+gamma*gamma);
      else
        extra = (double)projectileZ*pow((double)projectileZ/(double)targetZ,1./3.)*1.273e-2*gamma/(pow(epsi-3.458e-2,2)+gamma*gamma);

      switch(stragType) {
      case 0: // no straggling
      default:
        str = 0.;
        break;
      case 1: // Bohr's
        str = omegaBohr2;
        break;
      case 2: // Chu correction only
        str = omegaBohr2*chucorr;
        break;
      case 3: // Chu + Wang correction
        str = omegaBohr2*(chargestate*chucorr+extra);
      }
      sstrag += str*elementFraction[ilayer][i]; // Bragg's rule;
    }
    strag[k] = sqrtdf(sstrag)*sqrtdf(2.);
  }
  return strag;
}



int fillTables(ion_params *ion, float compoundCorr, unsigned int enerStragType) {
  int i1, i2;

  if(ion==NULL) return 2;
  if(ion->tablesComputed) return 1;

  for(i1=0; i1<=MAXLAYERS; i1++) {
    for(i2=0; i2<MAXELEMENT; i2++) {
      if(element[i1][i2]) {

        // screaning length
        //ion->cp[i1][i2].a = d2f(SCREENCONST/(pow(element[i1][i2]->Z,0.23)+pow(ion->Z,  0.23)));
        ion->a[i1][i2] = d2f(SCREENCONST/(pow(element[i1][i2]->Z,0.23)+pow(ion->Z,  0.23)));

        // mass ratio
        //ion->cp[i1][i2].mr = ion->mass/element[i1][i2]->mass;  // m1/m2 
        ion->mr[i1][i2] = ion->mass/element[i1][i2]->mass;  // m1/m2 
      
        // mass part of the kinematic factor
        //ion->cp[i1][i2].K = 4.0f*ion->cp[i1][i2].mr/((1.0f+ion->cp[i1][i2].mr)*(1.0f+ion->cp[i1][i2].mr));
        ion->K[i1][i2] = 4.0f*ion->mr[i1][i2]/((1.0f+ion->mr[i1][i2])*(1.0f+ion->mr[i1][i2]));

        // reduced energy conversion factor
        //ion->cp[i1][i2].f  = ion->cp[i1][i2].a/((ion->cp[i1][i2].mr+1.0f) * element[i1][i2]->Z*ion->Z *E2);
        ion->f[i1][i2]  = ion->a[i1][i2]/((ion->mr[i1][i2]+1.0f) * element[i1][i2]->Z*ion->Z *E2);

        // cos and sin components of Matrix  
        //ion->cp[i1][i2].CosScat = (float *) malloc(DIME*DIMS*sizeof(float));
        //ion->cp[i1][i2].SinScat = (float *) malloc(DIME*DIMS*sizeof(float));
        ion->CosScat[i1][i2] = (float *) malloc(DIME*DIMS*sizeof(float));
        ion->SinScat[i1][i2] = (float *) malloc(DIME*DIMS*sizeof(float));

        //if(ion->cp[i1][i2].CosScat==NULL || ion->cp[i1][i2].SinScat == NULL) {
        if(ion->CosScat[i1][i2]==NULL || ion->SinScat[i1][i2] == NULL) {
         fprintf(stderr, "Matrices memory allocation error.\n");
         exit(1);
        }  
        //fillCosSinTable(ion->cp[i1][i2].CosScat, ion->cp[i1][i2].SinScat, ion->cp[i1][i2].mr);
        fillCosSinTable(ion->CosScat[i1][i2], ion->SinScat[i1][i2], ion->mr[i1][i2]);

        // maximum reduced impact parameter
        //ion->cp[i1][i2].smax = 1.0f/(ion->cp[i1][i2].a*sqrtdf(PI*layerDensity[i1])*sqrtMeanFreePath);
        ion->smax[i1][i2] = 1.0f/(ion->a[i1][i2]*sqrtdf(PI*layerDensity[i1])*sqrtMeanFreePath[i1]);
      }
    }  

    if(numberElements[i1]) {
      // fill stopping power tables for this layer
      ion->sp[i1] = fillDedxTable(i1, ion->Z, ion->mass, layerDensity[i1], compoundCorr );
      if(!ion->sp[i1]) {
        fprintf(stderr, "\nError: can't load stopping power table for Z=%u\n", ion->Z);
        exit(1);
      }  

      // fill straggling table for this layer
      ion->sqrtStrag[i1] = fillStragglingTable(i1, ion->Z, ion->sp[i1], hsp[i1], enerStragType);
      if(!ion->sp[i1]) {
       exit(1);
      } 
    }
  }
  return 0;
}
  

// generate and randomize lists randomlist,sqrtloglist, sinAzimAngle & cosAzimAngle  
void computelists() {
  unsigned int irandomlist, iloglist, iazimAngle;
  // generate a uniformly spaced list of values between .5/MAXRANLIST and 1-.5/MAXRANLIST
  for(irandomlist=0; irandomlist<MAXRANLIST; irandomlist++)
    randomlist[irandomlist] = (irandomlist+0.5f)/MAXRANLIST;
  randomizelist(randomlist, MAXRANLIST); // put the list in random order
  for(irandomlist=0; irandomlist<MAXRANLIST; irandomlist++)
    sqrtrandomlist[irandomlist] = sqrtdf(randomlist[irandomlist]); // also compute sqrt of these values
 
  // produce a list of MAXLOGLIST sqrt(-log(x)) values for x uniformly distributed 
  // ...between x=.5/MAXLOGLIST and x=1-.5/MAXLOGLIST
  for(iloglist=0; iloglist<MAXLOGLIST; iloglist++)
    sqrtloglist[iloglist] = sqrtdf(-log((iloglist+.5)/MAXLOGLIST));
  randomizelist(sqrtloglist, MAXLOGLIST); // put the list in random order
  // precompute 1/sqrtloglist[]
  for(iloglist=0; iloglist<MAXLOGLIST; iloglist++)
    sqrtloglist1[iloglist] = 1.0f/sqrtloglist[iloglist];
 
  // produce a list uniformly distributed but randomly ordered azimutal angles  
  for(iazimAngle=0; iazimAngle<MAXAZILIST; iazimAngle++)
    // cosAzimAngle temporarly contains angles
    cosAzimAngle[iazimAngle]= 2.0f*(float)PI*(float)(iazimAngle)/(float)(MAXAZILIST);
  randomizelist(cosAzimAngle, MAXAZILIST); // put the list in random order
  for(iazimAngle=0; iazimAngle<MAXAZILIST; iazimAngle++) {
    // compute the cos and sine of these angles
    sinAzimAngle[iazimAngle] = d2f(sin(cosAzimAngle[iazimAngle]));
    cosAzimAngle[iazimAngle] = d2f(cos(cosAzimAngle[iazimAngle]));
  }

}


// Compute new directional cosines from cos & sin components of scattering angle theta
// and a random azimutal angle (no trigonometric functions called!)
//
// **** this function is critical in terms of computation efficiency ****
//
void rotate(float *l, float *m, float *n, unsigned int * iazimAngle, float costheta, float sintheta) {
  
  float k, kk, kinv;
  float ll=*l, mm=*m, nn=*n;
  float k2 = 1.0f-nn*nn;

  // random azimutal rotation angle components
  float cosomega = cosAzimAngle[*iazimAngle];
  float sinomega = sinAzimAngle[*iazimAngle];
  if(++*iazimAngle==MAXAZILIST) *iazimAngle = 0;

  if(k2==0) {
    // extreemly rare case
    *n = costheta;
    *m = sintheta*cosomega;
    *l = sintheta*sinomega;
    return;
  } 

  // usual case
  kinv = myInvSqrt(k2);  // 1/sqrt() (approximate)
  k = k2*kinv;           // so we get k and 1/k by a table lookup and a multiplication 

  // these two lines can be replaced by the following two lines but...
  // k  = sqrtdf(k2);    // ...using a sqrt() here makes the program 25% slower!
  // kinv = 1.0f/k;

  kk = sintheta*kinv;
  *l = ll * costheta+kk*(ll*nn*cosomega+mm*sinomega);
  *m = mm * costheta+kk*(mm*nn*cosomega-ll*sinomega);
  *n = nn * costheta-k*sintheta*cosomega;

}


// Compute ion slowdown to 'depth' in the target, taking into account stopping power, energy straggling and multiple collisions.
// *x is the depth (*x = 0 at surface and *x>0 inside the target)
// For an incomming particle, *x < depth and its initial value should be *x<0 but very small, e.g. *x = -1e-10
// Conversely, for an outgoing particle, depth < *x and depth should be < 0 but very small, e.g. depth = -1e-10.
// isTarget !=0 for particle in target, ==0 for particle in absorber/detector (layer numberLayers)
//
// **** this function is critical in terms of calculation efficiency ****
//
// Return:
//  0 for normal function end (depth>0 and *x>depth or depth<0 and *x<0)
//  1 if particle exited at bottom of target (e.g. outgoing particle reached target bottom)
// -1 if particle abnormally exited at top (e.g. incoming incident particle deflected outide target top)
// -2 if minimum energy is reached
int slowdown(float *x, float *y, float *z, float *l, float *m, float *n, float *E, 
             float depth, int isTarget, unsigned long *numcol, unsigned int *irandomlist, 
             unsigned int *iloglist, unsigned int *ierflist, unsigned int *iazimangle,
             ion_params *ion, int level) {

  int thisLevel = level+1;
  float fl, temp, temp2; // flight length, working variables
  float sqrtfl, ran, fraction;  // sqrt(fl), random number for collision partner, fraction of fl in current layer
  float S, sstr;       // computed stopping and straggling
  float ss;            // reduced impact parrameter, working variable

  unsigned int ilayer; // current layer; ilayer = numberLayers for absorber
  unsigned int ielem;  // collision partner

  unsigned long i;     // matrix index

  float sin2thetaby2;
  float amuRatio = MostAbundantIsotope[ion->Z]/ion->mass; // energy/velocity correction for stopping evaluation

  // these parameters are used when following recoils during implantation
  float Er, lr, mr, nr, xr, yr, zr; // recoil energy, and flying direction and position
  float lbc, mbc, nbc; // ion flying direction before collision (bc)
  float Erb, Edd, minE=minE0;
  ion_params * elem;
  unsigned int channel;
  int exitStatus;
  
  ilayer = isTarget?0:numberLayers; // index of current layer; ilayer = numberLayers for absorber

  while(*E>minE) {
    if(isTarget) { // if it's not target, it's absorber, and ilayer = numberLayers

      // find in which layer we are
      if(*x<layerTop[ilayer]) {
        do {
          if(ilayer==0) {
            return -1; // exit at top
          }
          ilayer--;
        } while(*x<layerTop[ilayer]);
      } else if(*x>=layerBottom[ilayer]) {
        do {
          ilayer++;
          if(ilayer>=numberLayers) {
            return 1; // exit at bottom
          }
        } while(*x>layerBottom[ilayer]);
      }
    }

    // randomly pick a partner for collision
    if(*irandomlist>=MAXRANLIST-2) *irandomlist = 0;
    ran = randomlist[(*irandomlist)++];
    for(ielem=0;ielem<numberElements[ilayer]; ielem++)
      if(elementFractionSum[ilayer][ielem]>=ran)
        break;

    // estimate sqrt of flight length according to Poisson's probability 
    sqrtfl  = sqrtloglist[*iloglist]*sqrtMeanFreePath[ilayer];
    fl = sqrtfl*sqrtfl;

    // pick an impact parameter (LSS units) (note that sqrtloglist1[*iloglist]==1/sqrtloglist[*iloglist])
    ss = sqrtrandomlist[(*irandomlist)++]*sqrtloglist1[*iloglist]* ion->smax[ilayer][ielem]; 
  //ss = sqrtdf(randomx())*smax/sqrtdf(-log(randomx())); // using this makes program 3 times slower!
    if(++*iloglist>=MAXLOGLIST) *iloglist=0;

    // readjust flight length so we don't go beyond depth
    temp = *l*depth;
    if(temp>0. // if particle is oriented towards depth (depth just smaller than 0 for surface)
      && temp<*l*(*x+*l*fl)) {  // ... and will go beyond depth... (multiplied by *l on each side of < so it works inward/outward)
      fl = (depth-*x)/ *l*1.0001f;     // readjust flight length so projectile flights slightly over depth
    }

    // electronic stopping
    i = Dindex(*E*amuRatio); // energy corrected for velocity in dE/dx calculation
    S = ion->sp[ilayer][i]*fl;

    // reajust stopping if we cross a layer boundary
    temp = *x+*l*fl;
    if(*l>0 && temp>layerBottom[ilayer]) { // bottom-oriented projectile crossing layer bottom
      fraction = (layerBottom[ilayer]-*x)/(*l*fl);
      S *= fraction; // fraction of stopping in this layer
      if(ilayer<numberLayers-1)
        S += ion->sp[ilayer+1][i]*(float)fl*(1.0f-fraction); // fraction in layer below, if any
    } else if(*l<0 && temp<layerTop[ilayer]) { // top-oriented projectile crossing layer top 
      fraction = (layerTop[ilayer]-*x)/(*l*fl);  // (l<0 and layerTop-x<0 also so signs cancel)
      S *= fraction; 
      if(ilayer>0 && ilayer<numberLayers-1)
        S += ion->sp[ilayer-1][i]*(float)fl*fraction; // fraction in layer above, if any (but absorber has one layer)
    }

    // advance to collision site
    *x = temp; // +=fl* *l;
    *y+=fl* *m;
    *z+=fl* *n;

    // compute energy straggling and subtract electronic energy loss
    sstr = ion->sqrtStrag[ilayer][i]*sqrtfl*erflist[(*ierflist)++];  // energy straggling; not reajusted, already a 1st order correction
    if(*ierflist>=MAXERFLIST) *ierflist=0; 
    *E -= S + sstr;

    // compute collision
    if(ss * ion->a[ilayer][ielem]>=interatomicDist[ilayer]) {
      // impact parameter larger than interatomic distance: assume collision has missed
      sin2thetaby2 = 0.0f;
      Er = 0;
    } else { // compute collision

      // get matrix index
      temp = *E* ion->f[ilayer][ielem];
      i = Eindex(temp)*DIMS+Sindex(ss);
      // energy transfer to recoil
      sin2thetaby2 = localmatrix[i]; //Matrix(i);
      Er = *E*ion->K[ilayer][ielem]*sin2thetaby2;

      if(followRecoils) {
        // store flying direction before collision 
        lbc = *l;
        mbc = *m;
        nbc = *n;
       
        // rotate ion trajectory
        rotate(l, m, n, iazimangle, ion->CosScat[ilayer][ielem][i], ion->SinScat[ilayer][ielem][i]);

        elem = element[ilayer][ielem];
        Edd = Ed[ilayer][ielem];
        if(Er>2.0f*Edd) {  // Er > 2Ed: generate a new recoil

          // compute recoil flying direction
          temp  = mySqrt(*E/Er * ion->mr[ilayer][ielem]);      // *********VERIFIER SI BON RAPPORT mr ***********
          temp2 = mySqrt((*E-Er)/ *E );
          lr = temp*(lbc - *l*temp2);
          mr = temp*(mbc - *m*temp2);
          nr = temp*(nbc - *n*temp2);

          xr = *x;
          yr = *y;
          zr = *z;

          // increment displacements histogram
          temp = (xr-zero)*invSlope+0.5f;
          channel = (unsigned int)temp;
          if(channel>=nchannels) channel = nchannels-1;
          dispHist[elem->Z][channel]++;

          // follow the generated recoil (recursive function call!)
          Erb = Er - Eb[ilayer][ielem]; // subtract binding energy
          exitStatus = slowdown(&xr, &yr, &zr, &lr, &mr, &nr, &Erb, depth, 1, 
                                 numcol, irandomlist, iloglist, ierflist, iazimangle, elem, thisLevel);

          switch(exitStatus) {
          case 1:
            exitBottom++;
            break;
          case -1:
            exitTop++;
            break;
          case -2:
            // minimum energy: normal function end for implanted ions and generated recoils
            temp = (xr-zero)*invSlope+0.5f;
            channel = (unsigned int)temp;
            if(channel>=nchannels) channel = nchannels-1;
            implHist[elem->Z][channel]++;
            break;
          default:
            // Unexpected but extreemly rare outcome due to a rotate() call with [l m n] = [0 0 1]  
            fprintf(stdout, "*");
            break;
          }

        } else if(Er>Edd) { // Ed < Er < 2Ed
          // ion has just enough energy to diplace the recoil, but we then assume that the ion stops here
          // (we wont't enter the replacement collision game)
          temp = (xr-zero)*invSlope+0.5f;
          channel = (unsigned int)temp;
          if(channel>=nchannels) channel = nchannels-1;
          dispHist[elem->Z][channel]++;
          implHist[elem->Z][channel]++;
        }
      } else {
        // not following recoils, just rotate ion trajectory
        rotate(l, m, n, iazimangle, ion->CosScat[ilayer][ielem][i], ion->SinScat[ilayer][ielem][i]);
      } 
      *E -= Er;
    }

    //fprintf(stderr, "\n%g %g %g %f %f %f %f %f %f %f %f",*E,sin2thetaby2,S,fl, sstr,*x,*y,*z,*l,*m,*n);
    if((depth>=0. && *x>=depth) || (depth<0. && *x<=0.)) {
      if(*E<minE) return -2; // last collision made energy lower than minimum energy
      return 0; // we have reach collision depth or surface (normal function end for IBA techniques)
    }
    (*numcol)++;

  }
  return -2; // we have reach minimum energy (normal function end for implantation)
}


// slowdown in a single, monoatomic layer
int quickslowdown(float *x, float *y, float *z, float *l, float *m, float *n, float *E, 
             float depth, int isTarget, unsigned long *numcol, unsigned int *irandomlist, 
             unsigned int *iloglist, unsigned int *ierflist, unsigned int *iazimangle,
             float amuRatio, float interDist, float top, float bottom, 
             float a, float f, float K, float smax, float * CosScat, float * SinScat, 
             float *sp, float *sqrtStrag, float sqrtMFP  ) {

  float fl, temp; // flight length, working variables
  float sqrtfl, fraction;  // sqrt(fl), random number for collision partner, fraction of fl in current layer
  float S, sstr;       // computed stopping and straggling
  float ss;            // reduced impact parrameter, working variable
  float Er = 0;

  unsigned long i;     // matrix index

  float sin2thetaby2;
  
  while(*E>minE0) {
      
    if(*x<top) {
      return -1; // exit at top
    } else if(*x>=bottom) { 
      return 1; // exit at bottom 
    }

    // estimate sqrt of flight length according to Poisson's probability 
    sqrtfl  = sqrtloglist[*iloglist]*sqrtMFP;
    fl = sqrtfl*sqrtfl;

    // readjust flight length so we don't go beyond depth
    temp = *l*depth;
    if(temp>0. // if particle is oriented towards depth (depth just smaller than 0 for surface)
      && temp<*l*(*x+*l*fl)) {  // ... and will go beyond depth... (multiplied by *l on each side of < so it works inward/outward)
      fl = (depth-*x)/ *l*1.0001f;     // readjust flight length so projectile flights slightly over depth
    }

    // electronic stopping
    i = Dindex(*E*amuRatio); // energy corrected for velocity in dE/dx calculation
    S = sp[i]*fl;

    // reajust stopping if we cross a layer boundary
    temp = *x+*l*fl;
    if(temp>bottom) { // bottom-oriented projectile crossing layer bottom
      fraction = (bottom-*x)/(*l*fl);
      S *= fraction; // fraction of stopping in this layer
    } else if(temp<top) { // top-oriented projectile crossing layer top 
      fraction = (top-*x)/(*l*fl);  // (l<0 and layerTop-x<0 also so signs cancel)
      S *= fraction; 
    }

    // advance to the next collision site
    *x = temp; // +=fl* *l;
    *y+=fl* *m;
    *z+=fl* *n;

    // energy straggling
    sstr = sqrtStrag[i]*sqrtfl*erflist[(*ierflist)++];  // energy straggling; not reajusted, already a 1st order correction
    if(*ierflist>=MAXERFLIST) *ierflist=0; 

    // subtract nuclear and electronic energy losses, including energy straggling
    *E -= S + sstr;

    // pick an impact parameter (LSS units) (note that sqrtloglist1[*iloglist]==1/sqrtloglist[*iloglist])
    ss = sqrtrandomlist[(*irandomlist)++]*sqrtloglist1[*iloglist]*smax; 
    if(++*iloglist>=MAXLOGLIST) *iloglist=0;
    if(*irandomlist>=MAXRANLIST) *irandomlist=0;

    if(ss*a>=interDist) {
      // impact parameter larger than interatomic distance: assume collision has missed
      sin2thetaby2 = 0.0f;
      Er = 0;

    } else {      // computer collision
      // compute matrix index
      i = Eindex(*E*f)*DIMS+Sindex(ss);

      // rotate directional cosines with precomputed cos and sin values of scattering angle
      sin2thetaby2 = localmatrix[i]; //Matrix(i);
      Er = *E*K*sin2thetaby2;       // recoil energy

      // not following recoils, just rotate ion trajectory
      rotate(l, m, n, iazimangle, CosScat[i], SinScat[i]);
    }

    // subtract nuclear and electronic energy losses, including energy straggling
    *E -= Er;

    //fprintf(stderr, "\n%g %g %g %f %f %f %f %f %f %f %f",*E,sin2thetaby2,S,fl, sstr,*x,*y,*z,*l,*m,*n);
    if((depth>=0. && *x>=depth) || (depth<0. && *x<=0.)) {
      if(*E<minE0) return -2; // last collision made energy lower than minimum energy
      return 0; // we have reach collision depth or surface (normal function end for IBA techniques)
    }
    (*numcol)++;
  }
  return -2; // we have reach minimum energy (normal function end for implantation)
}




// Find a new traveling direction within a cone of angle angSpread 
// around the line relating sample origin and detector "det" origin
void mainCollision(int det, float *cosx, float *cosy, float *cosz) {
  // Pic a random tilting angle uniformly distributed in terms of its cos component
  // (cosAngSpread randomly but uniformly distributed between cosAngSpread0 and 1)

  float cosAngSpread = (1.0f-cosAngSpread0)*(float)randomx()+cosAngSpread0; 
  float sinAngSpread = -sqrtdf(1-cosAngSpread*cosAngSpread);
  unsigned int iazim = (unsigned int)(MAXAZILIST*randomx());

  // new flying direction: direction of detector central point
  *cosx = pcnormx[det];
  *cosy = pcnormy[det];
  *cosz = pcnormz[det];

  // tilt this vector off its axis by an angle between angleSpread0 and angleSpread1
  // and rotate it by a random azimuthal angle
  rotate(cosx, cosy, cosz, &iazim, cosAngSpread, sinAngSpread);
}


// compute kinematic factor for scattering collision
// return 0 if collision forbiden by kinematics, 1 otherwise
// value of kinematic factor stored in *kinfact
int kinematicFactorRBS(
  float cosxin,  float cosyin,  float coszin,
  float cosxout, float cosyout, float coszout, 
	float *kinfact ) {

  float costhetaLab = cosxin*cosxout + cosyin*cosyout + coszin*coszout;
  float sin2thetaLab, mr2s2tl, sqrtfact, sqrtK;

  if(costhetaLab<=0.0f && massRatio>=1.0f) return 0;

  sin2thetaLab = 1.0f-costhetaLab*costhetaLab; // sin(theta_lab)^2
  mr2s2tl = massRatio*massRatio*sin2thetaLab;
  if(mr2s2tl>1.) {
    return 0; // scattering above critical angle
  } else {
    sqrtfact = mySqrt(1.0f-mr2s2tl);
    sqrtK = (sqrtfact+massRatio*costhetaLab)/(1.0f+massRatio);
    *kinfact = sqrtK*sqrtK;
    return 1;
  }
}


// compute kinematic factor for a recoil
// return 0 if collision forbiden by kinematics, 1 otherwise
// value of kinematic factor stored in *kinfact
int kinematicFactorERD(
  float cosxin,  float cosyin,  float coszin,
  float cosxout, float cosyout, float coszout, 
	float *kinfact ) {

  float costhetaLab = cosxin*cosxout + cosyin*cosyout + coszin*coszout;
  float temp;

	if(costhetaLab<0.) {
    return 0; // recoils only possible in forward direction
  } else {
    temp = 1.0f+massRatio;
    *kinfact = 4.0f*massRatio/(temp*temp)*costhetaLab*costhetaLab;
    return 1;
  }
}

// compute screened cross section of main collision (slow: do it only for detected particles)
float crosssectionRBS(int compScreen, float E,
               float cosxin, float cosyin, float coszin,
               float cosxout, float cosyout, float coszout,
               unsigned int beamZ, unsigned int recoilZ, float V1) {

  float costhetaLab = cosxin*cosxout + cosyin*cosyout + coszin*coszout;
  float sin2thetaLab = 1.0f-costhetaLab*costhetaLab; // sin(theta_lab)^2
  float sqrtfact, sqrtCrosssect, mr2s2tl, temp;

  // variables for computing screeing (default: no screening)
  float v1byEcm, thetaCM, sinthetaCMby2, screening = 1.0f;

  if(costhetaLab>cutoffCosthetaLab) {
    aboveCutoff++;
    return 0;
  }
  
  mr2s2tl = massRatio*massRatio*sin2thetaLab;

  // compute Andersen screening factor
  if(compScreen) {
    v1byEcm = V1/E*(massRatio+1.0f)*(massRatio+1.0f);

    // find center-of-mass angle
    // in case of RBS, thetaLab is the angle of the scattered particle
    thetaCM = d2f(acos(costhetaLab)-asin(-sqrt(sin2thetaLab)*massRatio));
    sinthetaCMby2 = d2f(sin(thetaCM*0.5));
  
    // compute Andersen screening factor
    temp = (1.0f+0.5f*v1byEcm);
    screening = temp*temp;
    temp = 0.5f*v1byEcm/sinthetaCMby2;
    temp = 1.0f+v1byEcm+temp*temp;
    screening *= 1.0f/(temp*temp);
  }
  
  if(mr2s2tl>1.) {
    return 0; // scattering above critical angle
  } else {
    sqrtfact = d2f(sqrt(1.0f-mr2s2tl));
    sqrtCrosssect = beamZ*recoilZ*E2/(2.0f*E)*(sqrtfact+costhetaLab)/sin2thetaLab;
    return sqrtCrosssect*sqrtCrosssect/sqrtfact*screening; // in A^2/str
  }
}

// compute screened cross section of main collision (slow: do it only for detected particles)
float crosssectionERD(int compScreen, float E,
               float cosxin, float cosyin, float coszin,
               float cosxout, float cosyout, float coszout,
               unsigned int beamZ, unsigned int recoilZ, float V1) {

  float costhetaLab = cosxin*cosxout + cosyin*cosyout + coszin*coszout;
  float sin2thetaLab = 1.0f-costhetaLab*costhetaLab; // sin(theta_lab)^2
  float sqrtCrosssect, mr2s2tl, temp;

  // variables for computing screeing (default: no screening)
  float v1byEcm, thetaCM, sinthetaCMby2,screening = 1.0f;
  
  mr2s2tl = massRatio*massRatio*sin2thetaLab;

  // compute Andersen screening factor
  if(compScreen) {
    v1byEcm = V1/E*(massRatio+1.0f)*(massRatio+1.0f);

    // find center-of-mass angle
    // in case of ERD, thetaLab is the angle of the recoil
    thetaCM = d2f(PI-2*acos(costhetaLab) );
    sinthetaCMby2 = d2f(sin(thetaCM*0.5));
  
    // compute Andersen screening factor
    temp = (1.0f+0.5f*v1byEcm);
    screening = temp*temp;
    temp = 0.5f*v1byEcm/sinthetaCMby2;
    temp = 1.0f+v1byEcm+temp*temp;
    screening *= 1.0f/(temp*temp);
  }
  
  if(costhetaLab<0.) {
    return 0; // recoils only possible in forward direction
  } else {
    temp = 1.0f+massRatio;
    sqrtCrosssect =  beamZ*recoilZ*E2*temp/(2.0f*E*costhetaLab);
    return  sqrtCrosssect*sqrtCrosssect/costhetaLab*screening; // in A^2/str
  }
}


// return squared distance to center if the trajectory defined by [x,y,z]+t[l,m,n] intersects detector, -1 otherwise
// i is the detector number to consider
// set [x,y,z] as the intersection point in sample reference frame
// set [xx yy zz] as the intersection point in detector reference frame
// set [ll mm nn] as the directional vector in detector reference frame
float intersection(int i, float x, float y, float z, float l, float m, float n, 
  float *xx, float *yy, float *zz, float *ll, float *mm, float *nn, float *t0, float winv, float hinv ) {
  float temp, temp1, temp2;
  float vx, vy, vz;

  *ll = l*xpx[i]+m*xpy[i]+n*xpz[i]; 
  *mm = l*ypx[i]+m*ypy[i]+n*ypz[i]; 
  *nn = l*ypx[i]+m*ypy[i]+n*ypz[i];

  *t0 = -(d[i]+xpx[i]*x+xpy[i]*y+xpz[i]*z)/(*ll); // t0=-(d-xp.r0)/(xp.u)
  if(*t0<0.0f)
    return -1; // particle directed away from detector

  // impact point r0+t0*u in sample reference frame
  x += l**t0;
  y += m**t0;
  z += n**t0;

  // vector from detector center to impact point in sample reference frame
  vx = x-pcx[i];
  vy = y-pcy[i];
  vz = z-pcz[i];
  temp = vx*vx+vy*vy+vz*vz; // v^2

  // impact point in detector reference frame
  *xx = 0.0f;                          // v.xp but v in plane and xp normal to plane
  *yy = vx*ypx[i]+vy*ypy[i]+vz*ypz[i]; // v.yp
  *zz = vx*zpx[i]+vy*zpy[i]+vz*zpz[i]; // v.zp

  temp1 = *yy*ainv[i]*winv;
  temp2 = *zz*binv[i]*hinv;
  if(isEllipse[i]) {
    // ellipse equation:          (v.yp/a)^2    + (v.zp/b)^2    = 1
    // virtual ellipse equation:  (v.yp/a Wf)^2 + (v.zp/b Hf)^2 = 1
    if( temp1*temp1+temp2*temp2 > 1.0f )  
      return -1; // particle outside ellipse
  } else {
    // rectangle
    // left of left side or right of right side or below bottom or above top
    if(   temp1<-1.0f    ||      temp1>1.0f     || temp2<-1.0f  || temp2>1.0f)
      return -1;
  }
  return temp;
}


int correctTrajectory(int det,  float *E, float Ebc,float Eac, float k, 
                      float *l, float *m, float *n, float lbc, float mbc, float nbc, float lac, float mac, float nac, 
                      float *x, float *y, float *z, float xc, float yc, float zc) {
  float temp, temp2;
  float cx, cy, cz;
  float costh, sinth;
  float lp, mp, np;
  float K;

  // positions in double precision because calculations involve the subtraction of nearly equal quantities
  double rvx, rvy, rvz;      // vector relating collision site and spot on virtual detector
  double rrx, rry, rrz;      // vector relating collision site and randomly selected spot on detector
  double xx = *x, yy = *y, zz = *z, kk;  // position in double
  double tempd, tempd2;

  int validKinematic;

  // let rv be a vector relating the collision site and virtual detector intersection point
  rvx = k* *l+xx-xc; 
  rvy = k* *m+yy-yc;
  rvz = k* *n+zz-zc;

  // let rr be the vector between collision site and the rescaled intersection point (in sample reference frame)
  // temporarly, rr = intersection point - pc
  rrx = k* *l+xx-pcx[det];
  rry = k* *m+yy-pcy[det];
  rrz = k* *n+zz-pcz[det];
  
  // tempd, tempd2 are the rescaled y,z components 
  // of the impact point in the detector reference frame
  tempd  = (rrx*ypx[det]+rry*ypy[det]+rrz*ypz[det])/Wf;
  tempd2 = (rrx*zpx[det]+rry*zpy[det]+rrz*zpz[det])/Hf;

  // finally, rr = temp*y' + temp2*z' + pc - colllision site
  rrx = tempd*ypx[det]+tempd2*zpx[det]+pcx[det]-xc;
  rry = tempd*ypy[det]+tempd2*zpy[det]+pcy[det]-yc;
  rrz = tempd*ypz[det]+tempd2*zpz[det]+pcz[det]-zc;

  // c = (rr x rv)/|rr x rv|: normal vector perpendicular to plane defined by rr and rv
  // first, compute rr x rv
  cx = d2f(rry*rvz-rrz*rvy);
  cy = d2f(rrz*rvx-rrx*rvz);
  cz = d2f(rrx*rvy-rry*rvx);
  // normalisation
  temp = sqrtdf((cx*cx+cy*cy+cz*cz)); // |rr x rv| also required for sinth computation below
  temp2 = 1.0f/temp;
  cx*=temp2;
  cy*=temp2;
  cz*=temp2;

  // angle components for rotation from rv to rr
  temp2 = 1.0f/sqrtdf((rvx*rvx+rvy*rvy+rvz*rvz)*(rrx*rrx+rry*rry+rrz*rrz)); // 1/|rr||rv|
  costh = d2f(rvx*rrx+rvy*rry+rvz*rrz)*temp2;  // costh = rr.rv/|rr||rv|
  //sinth = sqrtdf(1.-costh*costh);
  sinth = temp*temp2; // sinth = |rr x rv|/|rr||rv|

  // correction to the directionnal cosines at the exit of main collision
  // considering that the particle now flies to the real detector
  lp = lac*costh + (mac*cz-nac*cy)*sinth;
  mp = mac*costh + (nac*cx-lac*cz)*sinth;
  np = nac*costh + (lac*cy-mac*cx)*sinth;

  // store corrected flying direction at the exit of main collision for later cross section computation
  lac = lp; mac = mp; nac = np;

  // recompute collision kinematic factor and cross section considering corrected directional cosines lp, mp, np
  switch(simulationType) {
  case TECH_RBS:
  case TECH_RBS_TOF:
    validKinematic = kinematicFactorRBS(lbc, mbc, nbc, lp, mp, np, &K);
    break;
  case TECH_ERD:
  case TECH_ERD_TOF:
    validKinematic = kinematicFactorERD(lbc, mbc, nbc, lp, mp, np, &K);
    break;
  default:
    fprintf(stderr, "Virtual detection not implemented for simulation type %i.\n", simulationType);
    exit(1);
    break;
  }
  if(!validKinematic) return 0;

  // correction to the flying direction at target exit
  lp = *l*costh + (*m*cz-*n*cy)*sinth;
  mp = *m*costh + (*n*cx-*l*cz)*sinth;
  np = *n*costh + (*l*cy-*m*cx)*sinth;

  // correction of the energy loss considering the new flying direction
  // (since more or less distance is traveled through the sample)

  // rv now defines the vector between main collision site and target exit point in original direction
  rvx = xx-xc; // x ~ 0
  rvy = yy-yc;
  rvz = zz-zc;

  // rr = rotation of rv 
  rrx = rvx*costh + (rvy*cz-rvz*cy)*sinth;
  rry = rvy*costh + (rvz*cx-rvx*cz)*sinth;
  rrz = rvz*costh + (rvx*cy-rvy*cx)*sinth;

  // find new exit point
  kk = -xc/rrx; // trajectory is k*rr+[xc,yc,zc] and intersects surface at x=0 => k*rrx+xc = 0
  *x = d2f(kk*rrx+xc); 
  *y = d2f(kk*rry+yc);
  *z = d2f(kk*rrz+zc);
  
  // corrected flying direction
  *l = lp;
  *m = mp;
  *n = np;

  // rr now defines the vector between main collision site and new target exit point in corrected direction
  rrx = *x-xc;
  rry = *y-yc;
  rrz = *z-zc;

  // correction to energy transfer during main collision   and   energy loss
  *E = Ebc*K-(Eac-*E)*mySqrt(d2f((rrx*rrx+rry*rry+rrz*rrz)/(rvx*rvx+rvy*rvy+rvz*rvz)));
  return 1;
}


// main loop (called using a structure so we can also call it using POSIX threads)
// For each ion, it calls functions to compute:
//  - slowdown of incident ion
//  - new direction after main collision, kinematic factor
//  - slowdown of scattered ion or recoil
// The function then :
//  - determines intersection with real or virtual detector
//  - for virtual detector, selects a spot on real detector and make trajectory corrections
//  - calls function to compute slowdown through absorber/foil, if any
//  - determines intersection second detector, if any
//  - if particle is detected, calls function to compute screened cross-section and increment histogram by this amount
void mainloop(void *arg) {

  thread_params *param =(thread_params *)arg;
  unsigned long numProjectile    = param->numProjectile;
  unsigned long totalnumbercoll  = 0;
  float * histogram              = param->histogram;

  // local variables
  unsigned long channel, iProjectile, numbercoll = 0, numbercoll2 = 0;
  int exitStatus, validKinematic;
  float E, EE;            // energy, energy of recoil in coincidence
  float x, y, z;
  float l, m, n;       // position, directional cosines
  float x2, y2, z2;
  float l2, m2, n2; // position, directional cosines of recoil in coincidence detection
  float xp, yp, zp;
  float lp, mp, np;       // ...same, detector coordinates
  float x2p, y2p, z2p;
  float l2p, m2p, n2p; 

  float crosssect;
  float V1 = d2f(48.73*beam.Z*recoil.Z*sqrt(pow(beam.Z,2./3.)+pow(recoil.Z,2./3.))); // parameter for cross-section sreening
  float temp = d2f(randomx()), temp2;
  double temp3;

  float WfInv = 1.0f/Wf, HfInv = 1.0f/Hf;

  // create and initialize counters so each thread has its own
  unsigned int ir = (unsigned int)(temp*MAXRANLIST); // random
  unsigned int il = (unsigned int)(temp*MAXLOGLIST); // log
  unsigned int ie = (unsigned int)(temp*MAXERFLIST); // erf 
  unsigned int ia = (unsigned int)(temp*MAXAZILIST); // azimulal angle

  float collDepth;         // collision depth
  float t0;                // parameter defining collision position

  // variables related to virtual detector calculations
  float xc, yc, zc;         // position of main collision
  float lbc, mbc, nbc, Ebc; // directional cosines and energy just before main collision
  float lac, mac, nac, Eac; // directional cosines and energy just after main collision
  float lac2, mac2, nac2;
  float cx, cy, cz;         
  float K;                  // kinematic factor 

  int isQuick = numberLayers==1 && numberElements[0]==1 && simulationType!=-2;
  
  for(iProjectile=0; iProjectile<numProjectile; iProjectile++) { // ions loop

    if((iProjectile+1)%(numProjectile/10)==0) {
      fprintf(stdout, ".");
      fflush(stdout);
    }

    if(ie>MAXERFLIST-6) ie=0; 

    // starting projectile energy
    E = E0;

    // starting at the origin to avoid single precision problems
    // beam size effects added afterwards
    x = 1e-20f;
    y = 0;
    z = 0;

    // starting directional cosines
    l = l0;
    m = m0;
    n = n0;

    lp = 99;
    l2p = 99;
    lac = 99;
    lac2 = 99;

    if(simulationType>0) {
      // follow projectile to predefined depth of main collision
      collDepth = layerTop[collisionLayer]+layerThickness[collisionLayer]*(float)iProjectile/(float)numProjectile+1e-20f;   
    } else {
      // we want to follow implanted ions wherever they go, down to minimum energy or until they exit target
      collDepth = 1e32f; // ... so put collison depth at infinity
    }
    
    // compute projectile slowdown
    numbercoll = 0;
    if(isQuick)
      exitStatus = quickslowdown(&x, &y, &z, &l, &m, &n, &E, collDepth, 1, &numbercoll, &ir, &il, &ie, &ia,
              MostAbundantIsotope[beam.Z]/beam.mass, interatomicDist[0], layerTop[0], layerBottom[0], 
              beam.a[0][0], beam.f[0][0], beam.K[0][0], beam.smax[0][0], beam.CosScat[0][0], beam.SinScat[0][0],
              beam.sp[0], beam.sqrtStrag[0], sqrtMeanFreePath[0]);
    else
      exitStatus = slowdown(&x, &y, &z, &l, &m, &n, &E, collDepth, 1, &numbercoll, &ir, &il, &ie, &ia, &beam,0);
    totalnumbercoll+=numbercoll;

    // write details
    if(writeDetails&1) {
#ifdef MAX_THREAD
      pthread_mutex_lock(&mut);
#endif
      fprintf(ofpdetails, "0 %lu %lu %g %g %g %g %g %g %g\n", iProjectile, numbercoll, E, x, y, z, l, m, n);
#ifdef MAX_THREAD
      pthread_mutex_unlock(&mut);
#endif
    }

    // for IBA techiques, compute main collision and slowdown of emerging ion(s)
    if(simulationType>0 && exitStatus==0) {

      // store direction and energy before main collision (bc)
      lbc = l; mbc = m; nbc = n; Ebc = E;
		  // ... and position of main collision 
      xc = x; yc = y; zc = z; 
      x2 = x; y2 = y; z2 = z; 

      // find new flying direction towards detector 0
      mainCollision(0, &l, &m, &n); 

      // store direction after main collision (ac)
      lac = l; mac = m; nac = n;

      // find new flying direction and compute kinematic factor
      switch(simulationType) {
      case TECH_RBS: //RBS
      case TECH_RBS_TOF: //RBS-TOF
        validKinematic = kinematicFactorRBS(lbc, mbc, nbc, l, m, n, &K);
        E*=K;  // apply kinematic factor
        Eac=E; // store energy after main collision
        break;
      case TECH_ERD: //ERDA
      case TECH_ERD_TOF: //ERD-TOF
        validKinematic = kinematicFactorERD(lbc, mbc, nbc, l, m, n, &K);
        E*=K; // apply kinematic factor
        Eac=E; // store energy after main collision
        break;
      case TECH_COINC: //coincidence (two detectors)
        // scattering
        validKinematic = kinematicFactorRBS(lbc, mbc, nbc, l, m, n, &K);
        E*=K;       // apply kinematic factor
        EE = Ebc-E; // recoil energy considering an elastic collision

        // we assume identical masses, i.e. recoiling at 90 deg. from scattering
        if(beam.mass != recoil.mass) {
          fprintf(stderr, "\nCoincidence for different masses not implemented.\n");
          exit(1);
        }
        // determine recoil direction: [l2 m2 n2] = [l m n]_ac x c
        // c = unit vector parallel to [l m n]_bc x [l m n]_ac
        cx = mbc*nac-nbc*mac;
        cy = nbc*lac-lbc*nac;
        cz = lbc*mac-mbc*lac;
        temp = myInvSqrt(cx*cx+cy*cy+cz*cz);
        cx*=temp;
        cy*=temp;
        cz*=temp;

        // [l2 m2 n2] = [l m n]_ac x c/|[l m n]_ac x c| 
        // but |[l m n]_ac x c|=1 because they are unit vectors perp. to each other
        l2 = mac*cz-nac*cy;
        m2 = nac*cx-lac*cz;
        n2 = lac*cy-mac*cx;
        lac2 = l2; mac2 = m2; nac2 = n2;
        break;
      default:
        fprintf(stderr, "\nSimulation type %d unknown.\n",simulationType);
        exit(1);
        break;
      }
   
      if(!validKinematic) {
        // we are not continuing, so add beam size to positions
        temp = ySpotSize*erflist[ie++];
        temp2 = zSpotSize*erflist[ie++]; 
        y+=temp;
        z+=temp2;
        y2+=temp;
        z2+=temp2;
        yc+=temp;
        zc+=temp2;
      } else {

      
        // write details
        if(writeDetails&1) {

#ifdef MAX_THREAD
          pthread_mutex_lock(&mut);
#endif
          switch(simulationType) {
          case TECH_COINC: // coincidence
            fprintf(ofpdetails, "-1 %lu %lu %g %g %g %g %g %g %g\n", iProjectile, numbercoll2, EE, x2, y2, z2, l2, m2, n2);
            // no break
          case TECH_RBS:
          case TECH_RBS_TOF:
          case TECH_ERD:
          case TECH_ERD_TOF:
              fprintf(ofpdetails, "1 %lu %lu %g %g %g %g %g %g %g\n", iProjectile, numbercoll, E, x, y, z, l, m, n);
              break;
          } 
#ifdef MAX_THREAD
          pthread_mutex_unlock(&mut);
#endif
      
        }
      
        // slowdown of outgoing ion(s)
        switch(simulationType) {
        case TECH_RBS: //RBS
        case TECH_RBS_TOF: //RBS-TOF
          // slowdown of scattered ion up to just above surface (-1e-10) or until something else happen
          exitStatus = slowdown(&x, &y, &z, &l, &m, &n, &E, -1e-10f, 1, &numbercoll, &ir, &il, &ie, &ia, &beam,0);
          totalnumbercoll+=numbercoll;
          break;
        case TECH_ERD: //ERDA
        case TECH_ERD_TOF: //ERD-TOF
          exitStatus = slowdown(&x, &y, &z, &l, &m, &n, &E, -1e-10f, 1, &numbercoll, &ir, &il, &ie, &ia, &recoil,0);
          totalnumbercoll+=numbercoll;
          break;
        case TECH_COINC: // coincidence: follow both scattered ion and recoil
          exitStatus = slowdown(&x, &y, &z, &l, &m, &n, &E, layerBottom[numberLayers-1], 1, 
            &numbercoll, &ir, &il, &ie, &ia, &beam,0);
          totalnumbercoll+=numbercoll;

          if(exitStatus==0) {// only follow recoil if scattered ion made it to sample back
            exitStatus = slowdown(&x2, &y2, &z2, &l2, &m2, &n2, &EE, layerBottom[numberLayers-1], 1, 
              &numbercoll2, &ir, &il, &ie, &ia, &recoil,0);
            totalnumbercoll+=numbercoll2;
          }
          break;
        default:
          fprintf(stderr, "\nSimulation type %d unknown.\n",simulationType);
          exit(1);
          break;
        }


        // we have finished crossing the target, add beam size to positions
        temp = ySpotSize*erflist[ie++];
        temp2 = zSpotSize*erflist[ie++]; 
        y+=temp;
        z+=temp2;
        y2+=temp;
        z2+=temp2;
        yc+=temp;
        zc+=temp2;

        // detection
        if(exitStatus==0) {
          switch(simulationType ) {
          case TECH_COINC: // annular detector coincidence
            // intersection of scattered ion with detector 0
            temp = intersection(0, x, y, z, l, m, n, &xp, &yp, &zp, &lp, &mp, &np, &t0,1,1); 
            if(temp<r2annular[0]) {
              exitStatus = -3;
              break; // scattered ion not detected: no coincidence, don't bother to check recoil
            }
            // intersection of recoil with detector "det"
            temp = intersection(1, x2, y2, z2, l2, m2, n2, &x2p, &y2p, &z2p, &l2p, &m2p, &n2p, &t0,1,1); 
            if(temp<r2annular[1]) {
              exitStatus = -3;
              break; // recoil not detected: no coincidence
            }
            exitStatus = 0;
            break;
          case TECH_RBS: // RBS
          case TECH_RBS_TOF: //RBS-TOF
          case TECH_ERD: // ERDA
          case TECH_ERD_TOF: // ERD-TOF
            // intersection with detector 0
            temp = intersection(0, x, y, z, l, m, n, &xp, &yp, &zp, &lp, &mp, &np, &t0, WfInv, HfInv); // detector 0
            if(temp<0) {
              exitStatus = -3;
              break; // ion not detected
            }
            if(Hf!=1.0f || Wf!=1.0f) { 
              // correct for virtual detection
              validKinematic = correctTrajectory(0, &E, Ebc, Eac, t0, 
                &l, &m, &n, lbc, mbc, nbc, lac, mac, nac, 
                &x, &y, &z, xc, yc, zc);
            }
            // recompute corrected impact point to see if it is within annulus
            temp = intersection(0, x, y, z, l, m, n, &xp, &yp, &zp, &lp, &mp, &np, &t0,1,1);
            if(!validKinematic || temp<r2annular[0]) {
              exitStatus = -5;
              break; // ion not detected
            }
            break;
          default:
            fprintf(stderr, "\nSimulation type %d unknown.\n",simulationType);
            exit(1);
            break;
          }
        }

        // compute cross section
        switch(simulationType) {
        case TECH_RBS:  
        case TECH_RBS_TOF:
        case TECH_COINC: 
            crosssect = crosssectionRBS(computeScreen, Ebc, lbc, mbc, nbc, lac, mac, nac, beam.Z, recoil.Z, V1);
            break;
        case TECH_ERD: // ERD
        case TECH_ERD_TOF: // ERD-TOF
            crosssect = crosssectionERD(computeScreen, Ebc, lbc, mbc, nbc, lac, mac, nac, beam.Z, recoil.Z, V1);
            break;
        default:
            fprintf(stderr, "\nNo cross-section computation possible for simulation type %i\n", simulationType);
            exit(1);
            break;
        }
      
        if(writeDetails) { // writeDetails tested here so it is not tested several time when no details

#ifdef MAX_THREAD
          pthread_mutex_lock(&mut);
#endif

          // write details
          if(writeDetails&1) {
            switch(simulationType) {
            case TECH_COINC: // coincidence
              fprintf(ofpdetails, "-2 %lu %lu %g %g %g %g %g %g %g\n", iProjectile, numbercoll2, EE, x2, y2, z2, l2, m2, n2);
              // no break
            case TECH_RBS:
            case TECH_RBS_TOF:
            case TECH_ERD:
            case TECH_ERD_TOF:
              fprintf(ofpdetails, "2 %lu %lu %g %g %g %g %g %g %g\n", iProjectile, numbercoll, E, x, y, z, l, m, n);
              break;
            } 
          } 

          // write angle with detector of outcomming particle
          if(writeDetails&2) { 
            switch(simulationType) { 
            case TECH_COINC:
              if(l2p>=-1. && l2p <=1.) {
                temp3 = (lac2>=-1. && lac2<=1.)?180.-acos(lac2)*180./PI:-99;
                fprintf(ofangle, "  %g %g %g %i\n", collDepth, 180.-acos(l2p)*180./PI, temp3, exitStatus );
              }
              // no break
            case TECH_RBS:
            case TECH_RBS_TOF:
            case TECH_ERD:
            case TECH_ERD_TOF:
              if(lp>=-1. && lp<=1.) 
                temp3 = (lac>=-1. && lac<=1.)?180.-acos(lac)*180./PI:-99;
                fprintf(ofangle, "%g %g %g %i\n", collDepth, 180.-acos(lp)*180./PI, temp3, exitStatus);
              break;
            } 
          } 

          // write detector intersection position of detected particles in detector reference frame
          if(exitStatus==0 && writeDetails&4) {
            switch(simulationType) {
            case TECH_COINC: // coincidence
              fprintf(ofdetect, "-2 %10.5e %10.5e %10.5e %10.5e %10.5e %10.5e %10.5e %10.5e\n", 
                EE, x2p, y2p, z2p, l2p, m2p, n2p, crosssect*factorNtOmegaByNumdetpart);
              // no break
            case TECH_RBS:
            case TECH_RBS_TOF: 
            case TECH_ERD:
            case TECH_ERD_TOF: 
              fprintf(ofdetect, "2 %10.5e %10.5e %10.5e %10.5e %10.5e %10.5e %10.5e %10.5e\n", 
                E,  xp,  yp,  zp,  lp,  mp,  np,  crosssect*factorNtOmegaByNumdetpart);
              break;
            }
          } 

#ifdef MAX_THREAD
          pthread_mutex_unlock(&mut);
#endif
        } 

        // if absorber present, compute slowdown through it
        if(absorberThick && exitStatus==0) {

          // because absorber is thin compared to detector distances
          // we just move current depth from 0 to a depth corresponfing 
          // to absorber thickness and compute slowdown through it
          x  = absorberThick;
          xp = absorberThick;

          // slowdown through absorber (outcome could be other than exitStatus==0)
          switch(simulationType) {
          case TECH_RBS: //RBS
          case TECH_RBS_TOF: //RBS-TOF
            exitStatus = slowdown(&xp, &yp, &zp, &lp, &mp, &np, &E,  
              -1e-10f, 0, &numbercoll, &ir, &il, &ie, &ia, &beam,0);
            totalnumbercoll+=numbercoll;
            break;
          case TECH_ERD: //ERD
          case TECH_ERD_TOF: //ERD-TOF
            exitStatus = slowdown(&xp, &yp, &zp, &lp, &mp, &np, &E,  
              -1e-10f, 0, &numbercoll, &ir, &il, &ie, &ia, &recoil,0);
            totalnumbercoll+=numbercoll;
            break;
          default: //including TECH_COINC
            fprintf(stderr,"\nNo absorber/foil allowed for simulation type %i",simulationType);
            exit(1);
            break;
          }
        
          // time of flight techniques: detector 1 is stop detector, positioned in reference frame of detector 0 
          if( exitStatus==0 && (simulationType==TECH_RBS_TOF || simulationType==TECH_ERD_TOF) ) { 
            // did the particle reach the final detector? If yes, t0 gives the distance traveled between the start and stop detector
            if( intersection(1, xp, yp, zp, lp, mp, np, &x, &y, &z, &l, &m, &n, &t0,1,1)<0 ) {
              exitStatus=-4;
            }
          } 
        } 
      }
    }

    // energy difference discrimination for coincidence
    if(EdiffDiscr && simulationType==TECH_COINC && exitStatus==0 && (E-EE)*(E-EE)>EdiffDiscr*EdiffDiscr)
        exitStatus=-4;
/*
// this is to print the angle an event hits the detector compared to the angle it left the main 
// collision site as a function of depth, in order to appreciate the effects of multiple scattering
#ifdef MAX_THREAD
          pthread_mutex_lock(&mut);
#endif

          if(writeDetails&2 && !exitStatus) { 
            switch(simulationType) { 
            case TECH_COINC:
              if(l2p>=-1. && l2p <=1.) {
                temp3 = (lac2>=-1. && lac2<=1.)?180.-acos(lac2)*180./PI:-99;
                fprintf(ofangle, "  %g %g %g %i\n", collDepth, 180.-acos(l2p)*180./PI, temp3, exitStatus );
              }
              // no break
            case TECH_RBS:
            case TECH_RBS_TOF:
            case TECH_ERD:
            case TECH_ERD_TOF:
              if(lp>=-1. && lp<=1.) 
                temp3 = (lac>=-1. && lac<=1.)?180.-acos(lac)*180./PI:-99;
                fprintf(ofangle, "%g %g %g %i\n", collDepth, 180.-acos(lp)*180./PI, temp3, exitStatus);
              break;
            } 
          } 

#ifdef MAX_THREAD
          pthread_mutex_unlock(&mut);
#endif
*/
    // compute histograms and sums
    switch (exitStatus) {
    case 0:
      if(simulationType<0) {
        // Unexpected but extreemly rare outcome due to a rotate() call with [l m n] = [0 0 1] that return infinite value  
        fprintf(stdout, "*");
      } else {
        if(validKinematic) {
   
          // compute channel
          switch(simulationType) {
          case TECH_RBS: // RBS
          case TECH_ERD: // ERD
            if(quad==0)
              temp = (E-zero)*invSlope+0.5f; // +0.5 so calibration defines the energy of the middle of a channel
            else
              temp = (-slope+sqrtdf(slope*slope-4.*quad*(zero-E)))/(2.0f*quad)+0.5f; 
            break;
          case TECH_RBS_TOF: // RBS-TOF
            if(E>0) {
              temp = sqrtdf(beam.mass/(2.0*E)*AMUbyE)*t0*1e9f; // time in ns (t0 is the distance traveled by the particle)
              temp = (temp-zero)*invSlope+0.5f; // +0.5 so calibration defines the energy of the middle of a channel
            } else {
              temp = -1;
            }
            break;
          case TECH_ERD_TOF: // ERD-TOF
            if(E>0) {
              // time in ns (t0 is the distance traveled by the particle between start and stop detector)
              // (factor 0.1: sqrt is inv. speed in [1/(m/s)], t0 in [A] so factor 1e-10*1e9 to get time [ns])
              temp = sqrtdf(recoil.mass/(2.0*E)*AMUbyE)*t0*0.1f; 
              temp = (temp-zero)*invSlope+0.5f; // +0.5 so calibration defines the energy of the middle of a channel
            } else {
              temp = -1;
            }
            break;
          case TECH_COINC: // coincidence : energy is the sum of scattered ion and recoil energy
            if(quad==0)
              temp = (E+EE-zero)*invSlope+0.5f; // +0.5 so calibration defines the energy of the middle of a channel
            else
              temp = (-slope+sqrtdf(slope*slope-4.*quad*(zero-E-EE)))/(2.0f*quad)+0.5f; 
            break;
          default:
            fprintf(stderr, "\nNo energy or time spectrum possible for simulation type %i\n", simulationType);
            exit(1);
            break;
          }

          if(temp>=0) {
            channel = (unsigned int)temp;
            histogram[channel<nchannels?channel:nchannels-1]+=crosssect;
            detectedParticles++;
          } else {
            exitMinEnergy++; // calibration caused energy to generate event below channel 0
          }
        } else {
          exitStatus = -6;
        }
      }
      break;
    case -3:
      exitUndetected++;
      break;
    case -4:
      exitUndetectedFinal++;
      break;
    case -5:
      badRescale++;
      break;
    case -6:
      badKinematics++;
      break;
    case 1:
      exitBottom++;
      break;
    case -1:
      exitTop++;
      break;
    case -2:
      if(simulationType<0) {
        // minimum energy: normal function end for implanted ions
        temp = (x-zero)*invSlope+0.5f;
        channel = (unsigned int)temp;
        histogram[channel<nchannels?channel:nchannels-1]++;
      }
      exitMinEnergy++;
      break;
    default:
      fprintf(stderr, "Exit status %i unknown.\n",exitStatus);
      exit(1);
      break;
    }

  } // end of ion loop

  param->totalnumbercoll = totalnumbercoll;

#ifdef MAX_THREAD
  pthread_exit(0);
#endif

}


// skip comment lines, i.e. lines beggining by anything else than a digit, +, -, '.' or space
void skipComments(FILE * ifp) {
  unsigned int ch;

  if(ftell(ifp)>0) ignoreline(ifp); // if something was already read from file, ignore the rest of current line (e.g. return)

  ch = fgetc(ifp); // probe frist character of new line
  while(!isdigit(ch) && ch!=' ' && ch!='+' && ch!='-' && ch!='.' && ch != EOF) {
    // if its not a digit, +, -, '.' or a space, its a comment, ignore line ..
    ignoreline(ifp);
    //... and probe first character of next line
    ch = fgetc(ifp);
  }
  ungetc(ch, ifp); // put back last character into stream so it is ready to read
}


// check if elem already exists in element[][] table
// it it does, assign element[ilayer][ielement] to that element and return 1
// if it doesn't, allocate a new element structure and return 0
int assignElement(unsigned int Z, float mass, unsigned int ilayer, unsigned int ielement) {
  unsigned int i1, i2;
  ion_params * ion;
  
  for(i1=0; i1<=MAXLAYERS; i1++) {
    for(i2=0; i2<MAXELEMENT; i2++) {
      if(element[i1][i2] != NULL && element[i1][i2]->Z==Z && element[i1][i2]->mass==mass) {
        // identical element found, point to the same structure
        element[ilayer][ielement] = element[i1][i2];
        return 1;
      }
    }
  }


  // not found, allocate a new structure
  element[ilayer][ielement] = malloc(sizeof(ion_params));

  // initialize newly created structure
  ion = element[ilayer][ielement];
  ion->Z    = Z;
  ion->mass = mass;
  ion->tablesComputed = 0;

  for(i1=0; i1<=MAXLAYERS; i1++) {
    for(i2=0; i2<MAXELEMENT; i2++) {
      ion->CosScat[i1][i2] = NULL;
      ion->SinScat[i1][i2] = NULL;
    }
    ion->sp[i1] = NULL;
    ion->sqrtStrag[MAXLAYERS+1]; // sqrt of energy straggling * sqrt(2) (computed according to Bragg's rule)
  }

  return 0;
}


// The main function reads the input file, compute several parameters, calls functions to compute tables, call mainloop, and writes spectrum
// convention: energies in eV, lengths in A, densities in at/A3 (=10^24 at/cm4)
int main( ){

  unsigned long numProjectile; // number of projectiles 
  unsigned int  enerStragType; // type of energy straggling computation: 0=none, 1=Brag, 2=Chu, 3=Wang
  float         beamRadius;    // sigma of a gaussian, radially simetric beam (0 to ignore)

  unsigned long totalnumbercoll=0; // number of multiple scattering collisions

  float      *histogram;       // accumulated histogram
  float       maxChannel;      // maximum spectrum value
  float       compoundCorrBeam, compoundCorrRecoil; // stopping correction factor for compounds
  float       meanFreePath;    // user set mean free path (0 to use interatomic distance)

  float      *gauss;
  unsigned long nmax;

  // coordinates of 3 points defining a detector in space
  float       p1x[NUMDET], p1y[NUMDET], p1z[NUMDET];
  float       p2x[NUMDET], p2y[NUMDET], p2z[NUMDET];
  float       p3x[NUMDET], p3y[NUMDET], p3z[NUMDET]; 

  // other working variables
  unsigned long i, i1, i2, Z;
  float sum, temp, temp2, mass;
  clock_t t1;

  // file handles
  FILE *ifp, *ofp;
  char *mode = "r", *modew = "w";
  char filename[1000];

  // structures to pass parameters to mainloop
#ifdef MAX_THREAD
  thread_params param[MAX_THREAD];
  pthread_t threads[MAX_THREAD];
  pthread_mutex_init (&mut, NULL);
#else
  thread_params param[1];
#endif

  ofpdetails = NULL;
  ofangle    = NULL;
  ofdetect   = NULL;

  // fill randomlist, sqrtloglist, sinAzimAngle and cosAzimAngle
  computelists();
  mySqrtTableFill();

  // load scatteing matrix
  if(!loadMatrix()) {
      fprintf(stderr, "Error loading scattering matrix.\n"); 
      return 1;
  }

  // load the list of Chu straggling correction parameters
  if(!loadChu()) {
    fprintf(stderr, "Error loading Chu parameters.\n");
    return 1;
  }

  // load a list of inverse error function values 
  if(!loadErfInv()) {
    fprintf(stderr, "Error loading inverse error function file.\n");
    return 1;
  }
  randomizelist(erflist, MAXERFLIST); // put the list in random order

  // load parameters from file corteo.in
  // (floating point values are read as strings then converted to avoid formating problems)
  ifp = fopen("corteo.in", mode);
  if (ifp == NULL) {
    fprintf(stderr, "Can't open input file corteo.in.\n");
    return 1;
  }

  // simulation parameters
  skipComments(ifp);
  fscanf(ifp, "%d %g %g %g %u %u %u %g", &simulationType, &meanFreePath, &cosAngSpread0, &minE0, 
                                      &enerStragType,  &writeDetails, &computeScreen, &cutoffCosthetaLab);

  cosAngSpread0 = d2f(cos(cosAngSpread0*PI/180.));
  cutoffCosthetaLab = d2f(cos(cutoffCosthetaLab*PI/180.));

  computeScreen = computeScreen!=0;  // non-zero value to compute screening
  if(enerStragType>3) {
    fprintf(stderr, "Energy straggling calculation type %u unknown.\n",enerStragType);
    return 1;
  }

  // beam parameters
  skipComments(ifp);
  fscanf(ifp, "%u %f %lu %g %g", &beam.Z, &beam.mass, &numProjectile, &E0, &beamRadius); 
  if( beam.mass == 0.0f) beam.mass = MostAbundantIsotope[beam.Z]; // if mass==0 use default
  
  // directional cosines of incident beam
  skipComments(ifp);
  fscanf(ifp, "%g %g %g", &l0, &m0, &n0);
  
  // normalize beam directinal cosines
  sum = sqrtdf(l0*l0+m0*m0+n0*n0);
  l0/=sum;
  m0/=sum;
  n0/=sum;

  // projection of beam size on surface
  ySpotSize = beamRadius/sqrtdf(1.0-m0*m0);
  zSpotSize = beamRadius/sqrtdf(1.0-n0*n0);

  // spectrum parameters
  skipComments(ifp);
  fscanf(ifp, "%u %g %g %g %g %g", &nchannels, &zero, &slope, &quad, &resol, &EdiffDiscr);
  invSlope = 1.0f/slope;


  // partner for main collision
  skipComments(ifp);
  fscanf(ifp, "%u %g", &recoil.Z, &recoil.mass);
  if(recoil.mass == 0.) recoil.mass = MostAbundantIsotope[recoil.Z];
  massRatio = beam.mass/recoil.mass;

  // layer in which main collision occurs, total number of layers
  skipComments(ifp);
  fscanf(ifp, "%u", &collisionLayer); 
  skipComments(ifp);
  fscanf(ifp, "%u", &numberLayers); 

  // bound checking
  if(numberLayers>MAXLAYERS || numberLayers<1) {
    fprintf(stderr, "Error: number of layers must be between 1 and %u.\n", MAXLAYERS);
    return 1;
  }
  if(collisionLayer>=numberLayers) { 
    fprintf(stderr, "Error: index of collision layer must be between 0 and %d\n", numberLayers-1);
    return 1;
  }

  // layer and foil description
  for(i1=0; i1<=MAXLAYERS; i1++)
    for(i2=0; i2<MAXELEMENT; i2++)
      element[i1][i2] = NULL;

  for(i1=0; i1<=numberLayers; i1++) {
    skipComments(ifp);
    fscanf(ifp, "%g %g %u %g %g", &layerThickness[i1], &layerDensity[i1], &numberElements[i1], &compoundCorrBeam, &compoundCorrRecoil ); 

    // check the value of compound correction factors
    if(compoundCorrBeam<.3f || compoundCorrBeam>2.f ) {
      fprintf(stderr, "\nCompound correction factor of %f for beam ions probably wrong in layer %i.\n", compoundCorrBeam, i1);
      return 1;
    }
    if(compoundCorrRecoil<.3 || compoundCorrRecoil>2) {
      fprintf(stderr, "\nCompound correction factor of %f for recoils probably wrong in layer %i.\n", compoundCorrRecoil, i1);
      return 1;
    }

    // bound checking
    if(numberElements[i1]>MAXELEMENT || numberElements[i1]<1) {
      if(i1<numberLayers)
        fprintf(stderr, "Error: number of elements in layer %u must be between 1 and %u.\n", i1, MAXELEMENT);
      else
        fprintf(stderr, "Error: number of elements in foil must be between 1 and %i.\n", MAXELEMENT);
      return 1;
    }

    // compute interatomic distance, and use it as mean free path unless it was set by user
    interatomicDist[i1] = d2f(pow(3./(4*PI*layerDensity[i1]),1./3.));
    if(meanFreePath==0) sqrtMeanFreePath[i1] = sqrtdf(interatomicDist[i1]);
    else sqrtMeanFreePath[i1] = sqrtdf(meanFreePath);

    // depth of layer top and bottom
    if(i1==0 || i1>=numberLayers) {  // top layer or foils
      layerTop[i1] = 0.;
      layerBottom[i1] = layerThickness[i1];
    } else {
      layerTop[i1] = layerBottom[i1-1];
      layerBottom[i1] = layerTop[i1] + layerThickness[i1];
    }

    sum = 0.;
    // for each element in the layer
    for(i2=0; i2<numberElements[i1]; i2++) {
      // Z, mass, fraction, displacement energy, binding energy 
      skipComments(ifp);
      fscanf(ifp, "%u %g %g %g %g", &Z, &mass, &elementFraction[i1][i2], &Ed[i1][i2], &Eb[i1][i2] ); 
      if(mass==0.) mass = AtomicMass[Z];  // default is average atomic mass
      sum += elementFraction[i1][i2];
      assignElement(Z, mass, i1,i2); // set pointer to already existing element structure or allocate a new one
    }

    // concentration normalization and fraction sum calculation
    for(i2=0; i2<numberElements[i1]; i2++) {
      elementFraction[i1][i2]/=sum; 
      if(i2==0) elementFractionSum[i1][i2]=elementFraction[i1][i2]; 
      else elementFractionSum[i1][i2]=elementFractionSum[i1][i2-1]+elementFraction[i1][i2]; 
    }

    // H stopping needed to find charge state in straggling calculation
    hsp[i1] = fillDedxTable(i1, 1, 1.007f, layerDensity[i1], 1.0f );
    if(!hsp[i1]) {
      fprintf(stderr, "\nError: can't load hydrogen stopping power table\n");
      return 1;
    }

  }

  // fill/alllocate tables depending on technique
  followRecoils = 0; // default value
  switch (simulationType) {
  case TECH_IMPREC:

    followRecoils = 1;

    for(i1=0; i1<NUMBERELEMENTS; i1++) {
      dispHist[i1] = NULL;
      implHist[i1] = NULL;
    }
    for(i1=0; i1<=MAXLAYERS; i1++) {
      for(i2=0; i2<MAXELEMENT; i2++) {
        if(fillTables(element[i1][i2], 1, enerStragType)==0) { // assume no compound correction for recoils during implantation
          if(dispHist[element[i1][i2]->Z]==NULL) {
            dispHist[element[i1][i2]->Z] = calloc(nchannels,sizeof(float)); // allocate and initialize to 0
            implHist[element[i1][i2]->Z] = calloc(nchannels,sizeof(float)); // allocate and initialize to 0
          }
        } 
      }
    }

  case TECH_ERD:
  case TECH_ERD_TOF:
  case TECH_COINC:
    fillTables(&recoil, compoundCorrRecoil, enerStragType);
    // NO BREAK
  case TECH_IMPLANT:
  case TECH_RBS:
  case TECH_RBS_TOF:
    fillTables(&beam, compoundCorrBeam, enerStragType);
    break;
  default:
    fprintf(stderr, "Simulation type %d unknown.\n");
    return 1;
  }

  // detector parameters
  if(simulationType>0) {
    skipComments(ifp);
    for(i=0;i<NUMDET;i++) {
      // is the dectector an ellipse (!=0) or a rectangle (==0)
      fscanf(ifp, "%i", &isEllipse[i]); 

      // coordinates of 3 points in space defining a rectangle containing the detector surface [A]
      fscanf(ifp, "%g %g %g", &p1x[i], &p1y[i], &p1z[i]);
      fscanf(ifp, "%g %g %g", &p2x[i], &p2y[i], &p2z[i]);
      fscanf(ifp, "%g %g %g", &p3x[i], &p3y[i], &p3z[i]);

      // width and height multiplying factor for virtual detector, annulus radius [A]
      fscanf(ifp, "%g", &r2annular[i]);
      r2annular[i]*=r2annular[i];

      // compute normal vectors defining reference frame axes of this detector
      // zp = (p2-p1)/|p2-p1|
      zpx[i] = p2x[i]-p1x[i]; 
      zpy[i] = p2y[i]-p1y[i]; 
      zpz[i] = p2z[i]-p1z[i]; 
      temp = sqrtdf(1.0/(zpx[i]*zpx[i]+zpy[i]*zpy[i]+zpz[i]*zpz[i]));
      binv[i] = 2.0f*temp; // 1/half-size of detector along detector z axis
      zpx[i] *= temp; 
      zpy[i] *= temp; 
      zpz[i] *= temp; 

      // yp = (p3-p2)/|p3-p2|
      ypx[i] = p3x[i]-p2x[i]; 
      ypy[i] = p3y[i]-p2y[i]; 
      ypz[i] = p3z[i]-p2z[i]; 
      temp = sqrtdf(1.0/(ypx[i]*ypx[i]+ypy[i]*ypy[i]+ypz[i]*ypz[i]));
      ainv[i] = 2.0f*temp; // 1/half-size of detector along detector y axis
      ypx[i] *= temp; 
      ypy[i] *= temp; 
      ypz[i] *= temp; 

      // check that zp and yp are perpendicular
      if(zpx[i]*ypx[i]+zpy[i]*ypy[i]+zpy[i]*ypy[i]>1e-6f) {
        fprintf(stderr,"\nError: coordinates of detector %i not properly defined.\n",i);
        return 1;
      }

      // xp = yp x zp
      xpx[i] = ypy[i]*zpz[i]-ypz[i]*zpy[i];
      xpy[i] = ypz[i]*zpx[i]-zpz[i]*ypx[i];
      xpz[i] = ypx[i]*zpy[i]-ypy[i]*zpx[i];

      // detector center
      pcx[i] = (p3x[i]+p1x[i])/2.0f;
      pcy[i] = (p3y[i]+p1y[i])/2.0f;
      pcz[i] = (p3z[i]+p1z[i])/2.0f;
   
      // compute d in detector plane equation xpx*x+xpy*y+xpz*z+d=0
      d[i] = -(xpx[i]*p1x[i]+xpy[i]*p1y[i]+xpz[i]*p1z[i]); // d=-xp.p1

      //  check if detector is pointing towards target
      if(pcx[i]/xpx[i]>0) {
        fprintf(stderr,"\nError: detector %i pointing away from target.\n",i);
        return 1;
      } 
      temp = 1.0f/sqrtdf(pcx[i]*pcx[i]+pcy[i]*pcy[i]+pcz[i]*pcz[i]);
      pcnormx[i] = pcx[i]*temp;
      pcnormy[i] = pcy[i]*temp;
      pcnormz[i] = pcz[i]*temp;

      // coordinates of points 1 and 3 in detector reference frame
      p1yp[i] = (p1x[i]-pcx[i])*ypx[i]+(p1y[i]-pcy[i])*ypy[i]+(p1z[i]-pcz[i])*ypz[i];
      p1zp[i] = (p1x[i]-pcx[i])*zpx[i]+(p1y[i]-pcy[i])*zpy[i]+(p1z[i]-pcz[i])*zpz[i];

      p3yp[i] = (p3x[i]-pcx[i])*ypx[i]+(p3y[i]-pcy[i])*ypy[i]+(p3z[i]-pcz[i])*ypz[i];
      p3zp[i] = (p3x[i]-pcx[i])*zpx[i]+(p3y[i]-pcy[i])*zpy[i]+(p3z[i]-pcz[i])*zpz[i];

    }

    // virtual detector enlargement factors
    skipComments(ifp);
    fscanf(ifp, "%g %g", &Wf, &Hf);
    if(simulationType==TECH_COINC && (Wf!=1.0f || Hf!=1.0f)) {
      fprintf(stderr, "Warning: virtual detection not implemented for coincidence. Factors set to 1.\n");
      Wf = 1.0f;
      Hf = 1.0f;
    }
  }

  // end of file input
  fclose(ifp);


// Compute factor by which the histogram is mltiplied: density*thickness*solidAngle/number_of_detected_particle, with  
//   number_of_detected_particle = numProjectile*solidAngle_virtual/cone_solid_angle, 
//   solidAngle_virtual = solidAngle*Wf*Hf, and cone_solid_angle = 2.*PI*(1.0f-cosAngSpread0)  
// The spectrum then needs (externally) to be multipllied by the number of incident ions q and the concentration
// of the detected element in the layer. See documentation for a more detailed explanation.
  factorNtOmegaByNumdetpart = d2f(layerThickness[collisionLayer]*layerDensity[collisionLayer]*2.*PI*(1.0f-cosAngSpread0)/numProjectile)/(Wf*Hf);


  // print parts of what has been read
#ifdef MAX_THREAD
  fprintf(stdout, "%i threads running.\n", MAX_THREAD);
#else
  fprintf(stdout, "1 thread running.\n");
#endif

  // bound checking
  switch (simulationType) {
  case TECH_IMPLANT:
  case TECH_IMPREC:
    // in this case, maxChannel is the maximum depth
    maxChannel = nchannels*slope;
    if(maxChannel<layerBottom[numberLayers-1])
      fprintf(stderr, "Warning: ranges exceeding %g A are stored in channel %d.\n", maxChannel, nchannels-1);
    break;
  case TECH_RBS:
  case TECH_ERD:
  case TECH_COINC:
    // in this case, maxChannel is the maximum energy
    maxChannel = nchannels*nchannels*quad+nchannels*slope+zero;
    if(maxChannel<E0)
      fprintf(stderr, "Warning: Events exceeding %7g eV are stored in channel %d.\n", maxChannel, nchannels-1);
    break;
  case TECH_RBS_TOF:
    // in this case, maxChannel is the maximum time (ns)
    maxChannel = nchannels*slope+zero;
    if(maxChannel<sqrtdf(beam.mass/(2.*minE0)*AMUbyE)*1e9f)
      fprintf(stderr, "Warning: Events with time of flight larger than %7g ns are stored in channel %d.\n", maxChannel, nchannels-1);
  case TECH_ERD_TOF:
    // in this case, maxChannel is the maximum time (ns)
    maxChannel = nchannels*slope+zero;
    if(maxChannel<sqrtdf(recoil.mass/(2.*minE0)*AMUbyE)*1e9f)
      fprintf(stderr, "Warning: Events with time of flight larger than %7g ns are stored in channel %d.\n", maxChannel, nchannels-1);
    break;
  default:
    fprintf(stderr, "Simulation type %d unknown.\n");
    return 1;
  }

  // memory allocation for energy spectrum
  histogram = calloc(nchannels, sizeof(float)); // memory for spectrum, initialized to 0
  if(histogram==NULL) {
    fprintf(stderr, "Memory allocation error: can't create a specturm containing %u bins.\n", nchannels);
    return 1;
  }  

  switch (simulationType) {
  case TECH_IMPLANT:
    fprintf(stdout, "Simulating implantation of ");
    break;
  case TECH_IMPREC:
    fprintf(stdout, "Simulating full cascade of ");
    break;
  case TECH_RBS:
    fprintf(stdout, "Simulating RBS spectrum of ");
    break;
  case TECH_RBS_TOF:
    fprintf(stdout, "Simulating RBS-TOF spectrum of ");
    break;
  case TECH_ERD:
    fprintf(stdout, "Simulating ERD spectrum of ");
    break;
  case TECH_ERD_TOF:
    fprintf(stdout, "Simulating ERD-TOF spectrum of ");
    break;
  case TECH_COINC:
    fprintf(stdout, "Simulating 2-detectors coincidence of ");
    break;
  default:
    fprintf(stderr, "Simulation type %d unknown.\n");
    return 1;
  }

  if(E0<1e3f)
     fprintf(stdout, "%.1f eV ",E0);
  else if(E0<1e6f)
     fprintf(stdout, "%.1f keV ",E0/1e3f);
  else
     fprintf(stdout, "%.1f MeV ",E0/1e6f);

  if(simulationType>0) {
    fprintf(stdout, "%.0f%s->%.0f%s in layer %u\n",beam.mass,AtomicNames[beam.Z],recoil.mass,AtomicNames[recoil.Z],collisionLayer);
    fprintf(stdout, "Beam incident at %.1f deg. and detector at %.1f deg. from surface normal.\n",acos(l0)*180./PI,acos(xpx[0])*180./PI);
    if(simulationType==TECH_COINC) 
      fprintf(stdout, "  Second detector at %.1f deg. from surface normal.\n",acos(xpx[1])*180./PI);
    fprintf(stdout, "Beam and detector %.1f deg. apart.\n",acos(l0*xpx[0]+m0*xpy[0]+n0*xpz[0])*180./PI);
    if(simulationType==TECH_COINC) 
      fprintf(stdout, "  Beam and second detector %.1f deg. apart.\n",acos(l0*xpx[1]+m0*xpy[1]+n0*xpz[1])*180./PI);
  } else {
    fprintf(stdout, "%.0f%s\n",beam.mass,AtomicNames[beam.Z]);
    fprintf(stdout, "Beam incident at %.1f deg. from surface normal.\n",acos(l0)*180./PI);
  }

  // print layer composition
  for(i1=0; i1<numberLayers; i1++) {
    fprintf(stdout, "\nLayer %u: %6.0f A of", i1, layerThickness[i1]);
    for(i2=0; i2<numberElements[i1]; i2++) {
      if(elementFraction[i1][i2]>=0.01f)
        fprintf(stdout, " %3.0f%2s (%5.1f%%)",element[i1][i2]->mass,AtomicNames[element[i1][i2]->Z], elementFraction[i1][i2]*100); 
      else
        fprintf(stdout, " %3.0f%2s (%5g%%)",element[i1][i2]->mass,AtomicNames[element[i1][i2]->Z], elementFraction[i1][i2]*100); 
    }
    fprintf(stdout, ", N=%3.1fe22/cm3, Nt=%0.fe15/cm2", layerDensity[i1]*100, layerThickness[i1]*layerDensity[i1]*10.);
  }

  // last layer is absorber; thickness = 0 for no absorber
  absorberThick = layerThickness[numberLayers]; 
  if(simulationType>0) {
    fprintf(stdout, "\nAbsorber: ");
    if(absorberThick) {
      fprintf(stdout, "%.0f A of ", absorberThick);
      for(i2=0; i2<numberElements[numberLayers]; i2++) 
        fprintf(stdout, "%.0f%s (%g%%) ",element[numberLayers][i2]->mass,AtomicNames[element[numberLayers][i2]->Z], elementFraction[numberLayers][i2]*100); 
    } else
      fprintf(stdout, "not present.");
  }

  // open required detail files
  if(writeDetails&1) ofpdetails = fopen("corteo.details", modew);
  if(simulationType>0 && writeDetails&2) ofangle = fopen("corteo.angle", modew);
  if(simulationType>0 && writeDetails&4) 
    ofdetect = fopen("corteo.detect", modew);

  EminErr=0; EmaxErr=0; 
  SminErr=0; SmaxErr=0; 
  DminErr=0; DmaxErr=0;

  fprintf(stdout, "\nSimulation running");
  fflush(stdout);
  t1=clock();

  // pass thread specific arguments and start mainloop (THREADS NOT FULLY TESTED)
#ifdef MAX_THREAD
  int nthread = MAX_THREAD;
  if(numProjectile<100) nthread=1;

  // start nthread threads
  for (i1=0; i1<nthread; i1++) { 
    param[i1].id=i1;
    param[i1].numProjectile   = numProjectile/nthread;
    param[i1].totalnumbercoll = 0;
    param[i1].histogram       = calloc(nchannels, sizeof(float)); // initialized to 0

    if(param[i1].histogram == NULL) {
      fprintf(stderr, "Histogram allocation error.\n");
      exit(1);
    }

    pthread_create(&threads[i1], NULL, (void *)&mainloop, (void *)(param+i1));
  }

  // wait for threads completion
  for (i1=0; i1<nthread; i1++)	{
    pthread_join(threads[i1],NULL);
  }

  // sum spectra and number of collisions
  totalnumbercoll = 0;
  for (i1=0; i1<nthread; i1++)	{
    totalnumbercoll+=param[i1].totalnumbercoll;
    for (i2=0; i2<nchannels; i2++) 
      histogram[i2]+=param[i1].histogram[i2];
  }
#else
  // no threads (or only one!)
  param[0].id=0;
  param[0].numProjectile   = numProjectile;
  param[0].totalnumbercoll = 0;
  param[0].histogram       = calloc(nchannels, sizeof(float)); // initialized to 0

  if(param[0].histogram == NULL) {
    fprintf(stderr, "Histogram allocation error.\n");
    exit(1);
  }

  mainloop(&param);

  totalnumbercoll=param[0].totalnumbercoll;
  for (i2=0; i2<nchannels; i2++) 
    histogram[i2]=param[0].histogram[i2];
#endif

  // simulation finished
  temp = (float)(clock()-t1)/(float)CLOCKS_PER_SEC;
  fprintf(stdout, "done in %.0f s, %.1f coll/ion, %.1f us/ion, %.1f ns/coll\n", temp, 
    (float)totalnumbercoll/(float)numProjectile, temp*1.e6/(float)numProjectile,temp*1.e9/(float)totalnumbercoll );

  // Multiply histogram by density*thickness*solidAngle/number_of_detected_particle, with  
  //   number_of_detected_particle = numProjectile*solidAngle_virtual/cone_solid_angle, 
  //   solidAngle_virtual = solidAngle*Wf*Hf, and cone_solid_angle = 2.*PI*(1.0f-cosAngSpread0)  
  // The spectrum then needs to be multipllied by the number of incident ions q and the concentration
  // of the detected element in the layer. See documentation for a more detailed explanation.
  if(simulationType>0) {
    for(i1=0;i1<nchannels;i1++) 
      histogram[i1] *= factorNtOmegaByNumdetpart; // sum(dsigma/dOmega) * N t Omega_cone/numProjectile

    // convolute histogram for detector resolution
    if(resol>0.0f) {
      resol /= 2.35482f; // convert FWHM to sigma
      nmax = (int)(5*resol/slope); // nmax = 5*sigma in terms of channels
      if(nmax>0) {

        // create one half of a gaussian
        gauss = calloc(nmax, sizeof(float)); // memory for convolution function

        gauss[0] = 1;
        sum = 1;
        temp = slope*slope/(2.0f*resol*resol);
        for(i=1;i<nmax;i++) {
          gauss[i] = d2f(exp(-(float)i*(float)i*temp));
          sum += 2.0f*gauss[i];
        }
        sum = 1.0f/sum;
        for(i=0;i<nmax;i++) gauss[i] *= sum; // normalize this half gaussian to 1/2 ...

        // copy current spectrum
        for(i1=0;i1<nchannels;i1++) param[0].histogram[i1] = histogram[i1];

        // carry out convolution
        for(i1=0;i1<nchannels;i1++) {
          histogram[i1] = param[0].histogram[i1]*gauss[0];
          for(i2=1;i2<nmax;i2++) { 
            // add both halfs
            if(i1+i2<nchannels) histogram[i1] += param[0].histogram[i1+i2]*gauss[i2]; 
            if((long)i1-(long)i2>=0) histogram[i1] += param[0].histogram[i1-i2]*gauss[i2];
          }
        }
      }
    }
    // result of summations
    fprintf(stdout, "%13lu total incident ions\n", numProjectile);
    fprintf(stdout, "%13lu detected atoms\n", detectedParticles);
    fprintf(stdout, "%13lu particles emerged at surface but were undetected\n", exitUndetected);
    fprintf(stdout, "%13lu main collisions below cutoff angle\n", aboveCutoff);
    fprintf(stdout, "%13lu particles entered detection system but were finally not detected\n", exitUndetectedFinal);
    fprintf(stdout, "%13lu incomming particles abnormally exited at top\n", exitTop);
    fprintf(stdout, "%13lu particles exited at target bottom\n", exitBottom);
    fprintf(stdout, "%13lu particles implanted or below channel 0\n", exitMinEnergy);
    fprintf(stdout, "%13lu particles generated in a direction not allowed by kinetics\n", badKinematics);
    fprintf(stdout, "%13lu particles not detected after correction for virtual detector\n\n", badRescale);

  } else {
    // divide the distribution by the number of projectiles 
    // and the slope, so we get a distribution in at/A/ion
    temp = 1.0f/(numProjectile*slope);
    for(i1=0;i1<nchannels;i1++) histogram[i1] *= temp;

    if(followRecoils) {
      for(i2=1;i2<NUMBERELEMENTS;i2++) {
        if(implHist[i2]) {
          for(i1=0;i1<nchannels;i1++) implHist[i2][i1] *= temp;
        }
        if(dispHist[i2]) {
          for(i1=0;i1<nchannels;i1++) dispHist[i2][i1] *= temp;
        }
      }
    }

    // compute moments, avoiding last channel
    // as it may contain ion implanted out of bound
    sum = 0;
    for(i1=0;i1<nchannels-1;i1++) sum+=histogram[i1];

    temp = 0; // mean range
    for(i1=0;i1<nchannels-1;i1++) temp+=i1*slope*histogram[i1];
    temp /= sum;

    temp2 = 0; // range straggling
    for(i1=0;i1<nchannels-1;i1++) temp2+=(i1*slope-temp)*(i1*slope-temp)*histogram[i1];
    temp2 /= sum;
    temp2 = sqrtdf(temp2);

    // output results
    fprintf(stdout, "%13lu total incident ions\n", numProjectile);
    fprintf(stdout, "%13lu implanted ions\n", exitMinEnergy);
    fprintf(stdout, "%13lu backscattered ions\n", exitTop);
    fprintf(stdout, "%13lu transmitted ions\n", exitBottom);
    fprintf(stdout, "mean range = %g A\n", temp);
    fprintf(stdout, "straggling = %g A\n", temp2);

    if(detectedParticles)
      fprintf(stderr, "%13lu detected atoms (no ion should be detected during inplantation).\n", detectedParticles);
  }

  // write histogram
  ofp = fopen("corteo.out", modew);
  if(ofp) {
    fprintf(ofp, "%u %u\n",nchannels, detectedParticles);
    for(i1=0;i1<nchannels;i1++) {
      fprintf(ofp, "%g %g\n", i1*slope+zero, histogram[i1]);  // energy is that of the middle of a channel
    }
    fclose(ofp);
  } else {
    fprintf(stderr, "\nError: unable to open corteo.out\n\n");
  }

  if(simulationType==-2) {
    for(i2=1;i2<NUMBERELEMENTS;i2++) {
      if(dispHist[i2]) {
        sprintf(filename, "disp%03i.out", i2);
        ofp = fopen(filename, modew);
        if(ofp) {
          for(i1=0;i1<nchannels;i1++) {
            fprintf(ofp, "%g %g\n", i1*slope+zero, dispHist[i2][i1]); 
          }
          fclose(ofp);
        }
        sprintf(filename, "impl%03i.out", i2);
        ofp = fopen(filename, modew);
        if(ofp) {
          for(i1=0;i1<nchannels;i1++) {
            fprintf(ofp, "%g %g\n", i1*slope+zero, implHist[i2][i1]); 
          }
          fclose(ofp);
        }
      }
    }
  }

  // list errors in index functions
  if(EminErr) fprintf(stderr, "%13lu reduced energies below minimum value\n", EminErr);
  if(EmaxErr) fprintf(stderr, "%13lu reduced energies above maximum value\n", EmaxErr);
  if(SminErr) fprintf(stderr, "%13lu reduced impact parameters below minimum value\n", SminErr);
  if(SmaxErr) fprintf(stderr, "%13lu reduced impact parameters above maximum value\n", SmaxErr);
  if(DminErr) fprintf(stderr, "%13lu stopping power energies below minimum value\n", DminErr);
  if(DmaxErr) fprintf(stderr, "%13lu stopping power energies above maximum value\n", DmaxErr);

  if(writeDetails&1) fclose(ofpdetails);
  if(simulationType>0 && writeDetails&2) fclose(ofangle);
  if(simulationType>0 && writeDetails&4) fclose(ofdetect);

  return 0;
}


